from typing import List

from typing_extensions import override

from mentat.command.command import Command, CommandArgument
from mentat.session_context import SESSION_CONTEXT


class UndoCommand(Command, command_name="undo"):
    @override
    async def apply(self, *args: str) -> None:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        code_file_manager = session_context.code_file_manager

        errors = code_file_manager.history.undo()
        if errors:
            stream.send("\n".join(errors), style="error")
        stream.send("Undo complete", style="success")

    @override
    @classmethod
    def arguments(cls) -> List[CommandArgument]:
        return []

    @override
    @classmethod
    def argument_autocompletions(cls, arguments: list[str], argument_position: int) -> list[str]:
        return []

    @override
    @classmethod
    def help_message(cls) -> str:
        return "Undo the last change made by Mentat."



from openai.types.chat import ChatCompletionSystemMessageParam

from mentat.parsers.file_edit import FileEdit
from mentat.session_context import SESSION_CONTEXT
from mentat.session_input import collect_user_input


async def get_user_feedback_on_edits(
    file_edits: list[FileEdit],
) -> tuple[list[FileEdit], bool]:
    session_context = SESSION_CONTEXT.get()
    stream = session_context.stream
    conversation = session_context.conversation

    stream.send(
        "Apply these changes? 'Y/n/i' or provide feedback.",
        style="input",
    )
    user_response_message = await collect_user_input()
    user_response = user_response_message.data

    need_user_request = True
    match user_response.lower():
        case "y" | "":
            edits_to_apply = file_edits
            conversation.add_message(
                ChatCompletionSystemMessageParam(role="system", content="User chose to apply all your changes.")
            )
        case "n":
            edits_to_apply = []
            conversation.add_message(
                ChatCompletionSystemMessageParam(
                    role="system",
                    content="User chose not to apply any of your changes.",
                )
            )
        case "i":
            edits_to_apply = await _user_filter_changes(file_edits)
            if len(edits_to_apply) > 0:
                conversation.add_message(
                    ChatCompletionSystemMessageParam(
                        role="system",
                        content="User chose to apply some of your changes.",
                    )
                )
            else:
                conversation.add_message(
                    ChatCompletionSystemMessageParam(
                        role="system",
                        content="User chose not to apply any of your changes.",
                    )
                )

        case _:
            need_user_request = False
            edits_to_apply = []
            conversation.add_message(
                ChatCompletionSystemMessageParam(
                    role="system",
                    content="User chose not to apply any of your changes.",
                )
            )
            conversation.add_user_message(user_response)
    return edits_to_apply, need_user_request


async def _user_filter_changes(file_edits: list[FileEdit]) -> list[FileEdit]:
    new_edits = list[FileEdit]()
    for file_edit in file_edits:
        if await file_edit.filter_replacements():
            new_edits.append(file_edit)

    return new_edits



import asyncio
import json
import logging
from json import JSONDecodeError
from pathlib import Path
from typing import AsyncIterator, Dict

from jsonschema import ValidationError, validate
from openai.types.chat import ChatCompletionChunk
from openai.types.chat.completion_create_params import ResponseFormat
from typing_extensions import override

from mentat.errors import ModelError
from mentat.llm_api_handler import chunk_to_lines
from mentat.parsers.file_edit import FileEdit, Replacement
from mentat.parsers.parser import ParsedLLMResponse, Parser
from mentat.parsers.streaming_printer import StreamingPrinter
from mentat.prompts.prompts import read_prompt
from mentat.session_context import SESSION_CONTEXT

json_parser_prompt_filename = Path("json_parser_prompt.txt")

comment_schema = {
    "type": "object",
    "properties": {"type": {"enum": ["comment"]}, "content": {"type": "string"}},
}

edit_schema = {
    "type": "object",
    "properties": {
        "type": {"enum": ["edit"]},
        "filename": {"type": "string"},
        "starting-line": {"type": "integer"},
        "ending-line": {"type": "integer"},
        "content": {"type": "string"},
    },
}

creation_schema = {
    "type": "object",
    "properties": {
        "type": {"enum": ["creation"]},
        "filename": {"type": "string"},
    },
}

deletion_schema = {
    "type": "object",
    "properties": {
        "type": {"enum": ["deletion"]},
        "filename": {"type": "string"},
    },
}

rename_schema = {
    "type": "object",
    "properties": {
        "type": {"enum": ["rename"]},
        "filename": {"type": "string"},
        "new-filename": {"type": "string"},
    },
}

output_schema = {
    "type": "object",
    "properties": {
        "content": {
            "type": "array",
            "items": {
                "anyOf": [
                    comment_schema,
                    edit_schema,
                    creation_schema,
                    deletion_schema,
                    rename_schema,
                ]
            },
        }
    },
}


class JsonParser(Parser):
    @override
    def get_system_prompt(self) -> str:
        return read_prompt(json_parser_prompt_filename)

    @override
    def response_format(self) -> ResponseFormat:
        return ResponseFormat(type="json_object")

    @override
    def line_number_starting_index(self) -> int:
        return 0

    @override
    async def stream_and_parse_llm_response(self, response: AsyncIterator[ChatCompletionChunk]) -> ParsedLLMResponse:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream

        printer = StreamingPrinter()
        printer_task = asyncio.create_task(printer.print_lines())
        message = ""
        conversation = ""
        rename_map: Dict[Path, Path] = {}
        async for chunk in response:
            if self.shutdown.is_set():
                printer.shutdown_printer()
                await printer_task
                stream.send("\n\nInterrupted by user. Using the response up to this point.")
                break

            for content in chunk_to_lines(chunk):
                if not content:
                    continue
                message += content
                printer.add_string(content, end="")
        else:
            # Only finish printing if we don't quit from ctrl-c
            printer.wrap_it_up()
            await printer_task
        logging.debug("LLM Response:")
        logging.debug(message)

        try:
            response_json = json.loads(message)
            validate(instance=response_json, schema=output_schema)
        except JSONDecodeError:
            # Should never happen with OpenAI's response_format set to json
            stream.send("Error processing model response: Invalid JSON", style="error")
            return ParsedLLMResponse(message, "", [])
        except ValidationError:
            stream.send("Error processing model response: Invalid format given", style="error")
            return ParsedLLMResponse(message, "", [])

        file_edits: Dict[Path, FileEdit] = {}
        for obj in response_json["content"]:
            filename = (session_context.cwd / obj.get("filename", "")).resolve()
            if filename in rename_map:
                filename = rename_map[filename]
            match obj["type"]:
                case "comment":
                    conversation += obj["content"]
                    fileedit = None
                case "edit":
                    fileedit = FileEdit(
                        filename,
                        [
                            Replacement(
                                obj["starting-line"] - 1,
                                obj["ending-line"] - 1,
                                obj["content"].split("\n"),
                            )
                        ],
                        False,
                        False,
                        None,
                    )
                case "creation":
                    fileedit = FileEdit(filename, [], True, False, None)
                case "deletion":
                    fileedit = FileEdit(filename, [], False, True, None)
                case "rename":
                    new_filename = session_context.cwd / obj["new-filename"]
                    fileedit = FileEdit(filename, [], False, False, new_filename)
                    rename_map[new_filename] = filename
                case _:
                    # Should never happen with JSON validation
                    raise ModelError("Invalid JSON type")
            if fileedit is None:
                continue
            if filename not in file_edits:
                file_edits[filename] = fileedit
            else:
                # TODO: Add merge function to fileedit
                old_fileedit = file_edits[filename]
                file_edits[filename] = FileEdit(
                    filename,
                    fileedit.replacements + old_fileedit.replacements,
                    fileedit.is_creation or old_fileedit.is_creation,
                    fileedit.is_deletion or old_fileedit.is_deletion,
                    (
                        fileedit.rename_file_path
                        if fileedit.rename_file_path is not None
                        else old_fileedit.rename_file_path
                    ),
                )

        return ParsedLLMResponse(
            message,
            conversation,
            [file_edit for file_edit in file_edits.values()],
        )



from pathlib import Path

from datasets import load_dataset, DatasetDict  # type: ignore

from mentat.sampler.sample import Sample


SWE_BENCH_SAMPLES_DIR = Path(__file__).parent / "benchmarks" / "swe_bench_samples"


def download_swe_benchmarks(split: str = "dev") -> list[dict[str, str]]:
    """3 splits are available: dev (225), test (2.29k), and train (19k)."""
    dataset: DatasetDict = load_dataset("princeton-nlp/SWE-bench", split=split)  # type: ignore
    dataset: list[dict[str, str]] = [dict(benchmark) for benchmark in dataset]
    return dataset


def get_swe_samples(split: str = "dev", max_benchmarks: int | None = None) -> list[Sample]:
    """Return a list of SWE-Bench samples.

    If missing, download, convert to Samples and save locally.
    """
    split_dir = SWE_BENCH_SAMPLES_DIR / split
    saved_benchmarks = list(split_dir.glob("*.json"))
    if not split_dir.exists() or max_benchmarks and len(saved_benchmarks) < max_benchmarks:
        print(f"Downloading {split} split from SWE-Bench...")
        split_dir.mkdir(parents=True, exist_ok=True)
        dataset = download_swe_benchmarks(split)
        samples = [Sample.from_swe_bench(benchmark) for benchmark in dataset]
        for sample in samples:
            sample.save(split_dir / f"{sample.id}.json")
    else:
        samples = [Sample.load(fname) for fname in saved_benchmarks]

    if max_benchmarks:
        samples = samples[:max_benchmarks]
    print(f"Selected {len(samples)} benchmarks from '{split}'")
    return samples



# Adapted from https://github.com/encode/broadcaster
from __future__ import annotations

import asyncio
from asyncio import CancelledError, Queue
from collections import defaultdict
from contextlib import contextmanager
from typing import Any, AsyncGenerator, Dict, Iterator, Set

import attr


# TODO: This entire file is way overengineered and overcomplicated, and vastly needs to be simplified and improved
@attr.define
class Event:
    channel: str = attr.field()
    message: Any = attr.field()


class Subscriber:
    def __init__(self, queue: Queue[Event | None]) -> None:
        self._queue = queue

    async def __aiter__(self) -> AsyncGenerator[Event, None]:
        try:
            while True:
                yield await self.get()
        except Unsubscribed:
            pass

    async def get(self) -> Event:
        item = await self._queue.get()
        if item is None:
            raise Unsubscribed()
        return item


class MemoryBackend:
    def __init__(self):
        self._subscribed: Set[str] = set()
        self._universally_subscribed: bool = False
        self._missed_events: defaultdict[str, list[Event]] = defaultdict(list)

    def connect(self) -> None:
        self._published: Queue[Event] = Queue()

    def disconnect(self) -> None:
        pass

    async def join(self) -> None:
        await self._published.join()

    def subscribe(self, channel: str) -> None:
        self._subscribed.add(channel)
        for event in self._missed_events[channel]:
            self.publish(event.channel, event.message)
        self._missed_events[channel].clear()

    def unsubscribe(self, channel: str) -> None:
        self._subscribed.remove(channel)

    def universal_subscribe(self) -> None:
        self._universally_subscribed = True
        for events in self._missed_events.values():
            for event in events:
                self.publish(event.channel, event.message)
        self._missed_events.clear()

    def universal_unsubscribe(self) -> None:
        self._universally_subscribed = False

    # Since there is no maximum queue size, use the synchronous version of this function instead
    async def publish_async(self, channel: str, message: Any) -> None:
        event = Event(channel=channel, message=message)
        if channel in self._subscribed or self._universally_subscribed:
            await self._published.put(event)
        else:
            self._missed_events[channel].append(event)

    def publish(self, channel: str, message: Any) -> None:
        event = Event(channel=channel, message=message)
        if channel in self._subscribed or self._universally_subscribed:
            self._published.put_nowait(event)
        else:
            self._missed_events[channel].append(event)

    async def next_published(self) -> Event:
        event = await self._published.get()
        self._published.task_done()
        return event


class Unsubscribed(Exception):
    pass


class Broadcast:
    def __init__(self):
        self._subscribers: Dict[str, set[Queue[Event | None]]] = {}
        self._universal_subscribers: set[Queue[Event | None]] = set()
        self._backend = MemoryBackend()

    def __enter__(self) -> Broadcast:
        self.connect()
        return self

    def __exit__(self, *args: Any, **kwargs: Any) -> None:
        self.disconnect()

    def connect(self) -> None:
        self._backend.connect()
        self._listener_task = asyncio.create_task(self._listener())

    def disconnect(self) -> None:
        if self._listener_task.done():
            try:
                self._listener_task.result()
            except CancelledError:
                pass
        else:
            self._listener_task.cancel()
        self._backend.disconnect()

    async def join(self) -> None:
        """Blocks until all events have been processed"""
        await self._backend.join()

    async def _listener(self) -> None:
        while True:
            event = await self._backend.next_published()
            for queue in self._subscribers.get(event.channel, set()):
                queue.put_nowait(event)
            for queue in self._universal_subscribers:
                queue.put_nowait(event)

    # Since there is no maximum queue size, use the synchronous version of this function instead
    async def publish_async(self, channel: str, message: Any) -> None:
        await self._backend.publish_async(channel, message)

    def publish(self, channel: str, message: Any) -> None:
        self._backend.publish(channel, message)

    @contextmanager
    def subscribe(self, channel: str) -> Iterator[Subscriber]:
        queue: Queue[Event | None] = Queue()

        if not self._subscribers.get(channel):
            self._backend.subscribe(channel)
            self._subscribers[channel] = set()

        self._subscribers[channel].add(queue)
        yield Subscriber(queue)
        self._subscribers[channel].remove(queue)

        if not self._subscribers.get(channel):
            del self._subscribers[channel]
            self._backend.unsubscribe(channel)

    @contextmanager
    def universal_subscribe(self) -> Iterator[Subscriber]:
        queue: Queue[Event | None] = Queue()

        if not self._universal_subscribers:
            self._backend.universal_subscribe()

        self._universal_subscribers.add(queue)
        yield Subscriber(queue)
        self._universal_subscribers.remove(queue)

        if not self._universal_subscribers:
            self._backend.universal_unsubscribe()



from pathlib import Path

from typing_extensions import override

from mentat.code_file_manager import CodeFileManager
from mentat.errors import ModelError
from mentat.parsers.change_display_helper import DisplayInformation, FileActionType
from mentat.parsers.file_edit import FileEdit, Replacement
from mentat.parsers.parser import ParsedLLMResponse, Parser
from mentat.prompts.prompts import read_prompt
from mentat.session_context import SESSION_CONTEXT

replacement_parser_prompt_filename = Path("replacement_parser_prompt.txt")


class ReplacementParser(Parser):
    @override
    def get_system_prompt(self) -> str:
        return read_prompt(replacement_parser_prompt_filename)

    @override
    def _could_be_special(self, cur_line: str) -> bool:
        return cur_line.startswith("@")

    @override
    def _starts_special(self, line: str) -> bool:
        return line.startswith("@") and len(line.strip()) > 1

    @override
    def _ends_special(self, line: str) -> bool:
        return line.startswith("@")

    @override
    def _special_block(
        self,
        code_file_manager: CodeFileManager,
        cwd: Path,
        rename_map: dict[Path, Path],
        special_block: str,
    ) -> tuple[DisplayInformation, FileEdit, bool]:
        info = special_block.strip().split(" ")[1:]
        if len(info) == 0:
            raise ModelError("Error: Invalid model output")

        file_name = Path(info[0])
        full_path = (cwd / file_name).resolve()
        file_lines = self._get_file_lines(code_file_manager, rename_map, full_path)
        new_name = None

        # For an insert, just make the second number 1 less than the starting line (since we sub 1 from starting line)
        if len(info) == 2 and info[1].startswith("insert_line="):
            info.append(f"ending_line={int(info[1].split('=')[1])-1}")

        if len(info) == 2:
            starting_line = 0
            ending_line = 0
            removed_lines = []
            match info[1]:
                case "+":
                    file_action_type = FileActionType.CreateFile
                case "-":
                    file_action_type = FileActionType.DeleteFile
                case _:
                    file_action_type = FileActionType.RenameFile
                    new_name = Path(info[1])
        elif len(info) == 3:
            try:
                # Convert from 1-index to 0-index
                starting_line = int(info[1].split("=")[1]) - 1
                # 1-index to 0-index = -1, inclusive to exclusive = +1
                ending_line = int(info[2].split("=")[1])
            except (ValueError, IndexError):
                raise ModelError("Error: Invalid line numbers given")
            if starting_line > ending_line or starting_line < 0 or ending_line < 0:
                raise ModelError("Error: Invalid line numbers given")
            removed_lines = file_lines[starting_line:ending_line]
            file_action_type = FileActionType.UpdateFile
        else:
            raise ModelError("Error: Invalid model output")

        display_information = DisplayInformation(
            file_name,
            file_lines,
            [],
            removed_lines,
            file_action_type,
            starting_line,
            ending_line,
            new_name=new_name,
        )

        file_edit = FileEdit(
            full_path,
            [],
            is_creation=file_action_type == FileActionType.CreateFile,
            is_deletion=file_action_type == FileActionType.DeleteFile,
            rename_file_path=(cwd / new_name).resolve() if new_name else None,
        )
        has_code = file_action_type == FileActionType.UpdateFile
        return (display_information, file_edit, has_code)

    @override
    def _ends_code(self, line: str) -> bool:
        return line.strip() == "@"

    @override
    def _add_code_block(
        self,
        code_file_manager: CodeFileManager,
        rename_map: dict[Path, Path],
        special_block: str,
        code_block: str,
        display_information: DisplayInformation,
        file_edit: FileEdit,
    ):
        file_edit.replacements.append(
            Replacement(
                display_information.first_changed_line,
                display_information.last_changed_line,
                code_block.split("\n")[:-2],
            )
        )

    def file_edits_to_llm_message(self, parsedLLMResponse: ParsedLLMResponse) -> str:
        """
        Inverse of stream_and_parse_llm_response
        """
        session_context = SESSION_CONTEXT.get()

        ans = parsedLLMResponse.conversation.strip() + "\n\n"
        for file_edit in parsedLLMResponse.file_edits:
            action_indicator = ""
            if file_edit.is_creation:
                action_indicator = "+"
            elif file_edit.is_deletion:
                action_indicator = "-"
            elif file_edit.rename_file_path is not None:
                action_indicator = file_edit.rename_file_path.relative_to(session_context.cwd).as_posix()

            file_rel_path = file_edit.file_path.relative_to(session_context.cwd).as_posix()
            if action_indicator != "":
                ans += f"@ {file_rel_path} {action_indicator}\n"

            for replacement in file_edit.replacements:
                if replacement.starting_line == replacement.ending_line and len(replacement.new_lines) != 0:
                    ans += f"@ {file_rel_path} insert_line={replacement.starting_line + 1}\n"
                else:
                    ans += (
                        f"@ {file_rel_path} starting_line={replacement.starting_line + 1} "
                        f"ending_line={replacement.ending_line}\n"
                    )

                if len(replacement.new_lines) != 0:
                    ans += "\n".join(replacement.new_lines) + "\n"
                ans += "@\n"

        return ans



import os
import shlex
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Set, TypedDict, cast

import attr
from pygments.lexer import Lexer
from pygments.lexers import guess_lexer_for_filename
from pygments.token import Token
from pygments.util import ClassNotFound

from mentat.command.command import Command
from mentat.session_context import SESSION_CONTEXT
from mentat.utils import get_relative_path

SECONDS_BETWEEN_REFRESH = 5

whitespace = " \t\r\n"


class Completion(TypedDict):
    """
    Represents a single unserialized auto-completion suggestion
    """

    # The completion to add to the buffer
    content: str
    # Number of characters from the cursor the completion should be placed (negative means backwards)
    position: int
    # The display name; if not given, will default to content
    display: str | None


@attr.define
class FileCompletion:
    last_updated: datetime = attr.field()
    syntax_fragments: Set[str] = attr.field()


def get_command_filename_completions(cur_path: str) -> List[str]:
    """
    Used by commands like include, exclude, and screenshot to get filename completions.
    """
    path = Path(cur_path)

    if cur_path.endswith("/") or cur_path.endswith(os.path.sep) or cur_path.endswith("~"):
        actual_parent = path
        cur_file = ""
    else:
        actual_parent = path.parent
        if not path.parts:
            cur_file = ""
        else:
            cur_file = path.parts[-1]
    parent_path = actual_parent.expanduser()
    if not parent_path.is_dir():
        return []

    completions: List[str] = []
    for child_path in parent_path.iterdir():
        child_file = child_path.parts[-1]
        # is_dir = child_path.is_dir()
        if str(child_file).startswith(cur_file):
            completions.append(
                # Because prompt_toolkit doesn't start a new completion until after the user types something,
                # best UI unfortunately seems to be not having a trailing / and having the user type it.
                str(actual_parent / child_file)  # + (os.path.sep if is_dir else "")
            )

    return completions


class AutoCompleter:
    def __init__(self):
        self._all_file_completions: Set[str] = set()
        self._file_completions: Dict[Path, FileCompletion] = dict()
        self._last_refresh_at: datetime | None = None

    def _replace_last_word(
        self,
        last_word: str,
        completions: List[tuple[str, str]],
        position: int | None = None,
    ) -> List[Completion]:
        """
        Takes a list of string completions along with their displays, and filters to match only completions
        whose displays start with the last word, and then returns a list of Completions that will replace that word
        """
        filtered_completions = [completion for completion in completions if completion[1].startswith(last_word)]
        return [
            Completion(
                content=completion[0],
                position=-len(last_word) if position is None else position,
                display=completion[1],
            )
            for completion in filtered_completions
        ]

    def _partial_shlex_split(self, argument_buffer: str) -> tuple[List[str], bool, bool]:
        """
        Handles unescaped backslashes and unclosed quotation marks
        """

        in_quote = False
        try:
            split_buffer = shlex.split(argument_buffer)
        except ValueError as e:
            if "quotation" in str(e):
                # Try to close with a single quotation mark; if that doesn't work, close with a double quotation mark
                try:
                    split_buffer = shlex.split(argument_buffer + '"')
                except ValueError as e:
                    split_buffer = shlex.split(argument_buffer + "'")
                in_quote = True
            elif "escaped" in str(e):
                # Remove the offending backslash and pretend it isn't there
                split_buffer, in_quote, _ = self._partial_shlex_split(argument_buffer[:-1])
                return split_buffer, in_quote, True
            else:
                raise ValueError(f"shlex.split raised unexpected error: {e}")
        return split_buffer, in_quote, False

    def _find_shlex_last_word_position(self, argument_buffer: str, num_words: int) -> int:
        """
        Find where the last word starts in a shlexed buffer
        """
        lex = shlex.shlex(argument_buffer, posix=True)
        lex.whitespace_split = True
        for _ in range(num_words - 1):
            lex.get_token()
        remaining = list(lex.instream)
        return 0 if not remaining else -len(remaining[0])

    def _command_argument_completion(self, buffer: str) -> List[Completion]:
        if any(buffer.startswith(space) for space in whitespace):
            return []
        if not any(space in buffer for space in whitespace):
            return self._replace_last_word(buffer, [(name, name) for name in Command.get_command_names()])
        else:
            command_cls = Command.create_command(buffer.split()[0]).__class__
            argument_buffer = buffer.split(maxsplit=1)
            if len(argument_buffer) < 2:
                argument_buffer = ""
            else:
                argument_buffer = argument_buffer[1]

            # shlex fails if there are uncompleted quotations or backslashes not escaping anything;
            # we check for this and try to complete the quotations/backslashes, so that we can get the
            # actual escaped last_word argument to compare against our possible completions.
            split_buffer, in_quote, unescaped = self._partial_shlex_split(argument_buffer)
            last_word_position = self._find_shlex_last_word_position(argument_buffer, len(split_buffer))

            # If we removed an ending \ and before it was whitespace, we need to move on one argument
            if unescaped and buffer[-2] in whitespace:
                split_buffer.append("")

            # shlex.split doesn't count the ending space
            if buffer[-1] in whitespace and not in_quote:
                split_buffer.append("")
                last_word_position = 0
            arg_position = len(split_buffer) - 1

            arg_completions = [
                (shlex.quote(name), name) for name in command_cls.argument_autocompletions(split_buffer, arg_position)
            ]
            return self._replace_last_word(split_buffer[-1], arg_completions, last_word_position)

    def _refresh_file_completion(self, file_path: Path):
        if not file_path.exists() or file_path.is_dir():
            if file_path in self._file_completions:
                del self._file_completions[file_path]
            return
        with open(file_path, "r") as f:
            file_content = f.read()

        try:
            lexer = cast(Lexer, guess_lexer_for_filename(file_path, file_content))
        except ClassNotFound:
            self._file_completions[file_path] = FileCompletion(datetime.utcnow(), set())
            return

        tokens = list(lexer.get_tokens(file_content))
        filtered_tokens = set[str]()
        for token_type, token_value in tokens:
            if token_type not in Token.Name:
                continue
            if len(token_value) <= 1:
                continue
            filtered_tokens.add(token_value)
        self._file_completions[file_path] = FileCompletion(datetime.utcnow(), filtered_tokens)

    def _refresh_all_file_completions(self):
        ctx = SESSION_CONTEXT.get()

        file_paths = ctx.code_context.include_files.keys()

        # Remove syntax completions for files not in the context
        for file_path in set(self._file_completions.keys()):
            if file_path not in file_paths:
                del self._file_completions[file_path]

        # Add/update syntax completions for files in the context
        for file_path in file_paths:
            if file_path not in self._file_completions:
                self._refresh_file_completion(file_path)
            else:
                modified_at = datetime.utcfromtimestamp(os.path.getmtime(file_path))
                if self._file_completions[file_path].last_updated < modified_at:
                    self._refresh_file_completion(file_path)

        # Build de-duped syntax completions
        self._all_file_completions = set[str]()
        for file_path, file_completion in self._file_completions.items():
            rel_path = get_relative_path(file_path, ctx.cwd)
            self._all_file_completions.add(str(rel_path))
            self._all_file_completions.update(file_completion.syntax_fragments)

        self._last_refresh_at = datetime.utcnow()

    def get_file_completions(self, buffer: str) -> List[Completion]:
        if (
            self._last_refresh_at is None
            or (datetime.utcnow() - self._last_refresh_at).seconds > SECONDS_BETWEEN_REFRESH
        ):
            self._refresh_all_file_completions()

        if not buffer or buffer[-1] == " ":
            return []

        words = buffer.split()
        if not words:
            return []

        last_word = words[-1]
        position = -len(last_word)
        if last_word.startswith("`"):
            last_word = last_word[1:]

        completions = [(f"`{completion}`", completion) for completion in self._all_file_completions]
        return self._replace_last_word(last_word, completions, position)

    def get_completions(self, buffer: str, command_autocomplete: bool = False) -> List[Completion]:
        """
        Get the auto-completion suggestions for the current user buffer.
        """
        if not buffer.strip():
            return []

        if buffer.startswith("/") and command_autocomplete:
            return self._command_argument_completion(buffer[1:])
        else:
            return self.get_file_completions(buffer)



from typing import List

from typing_extensions import override

from mentat.auto_completer import get_command_filename_completions
from mentat.command.command import Command, CommandArgument
from mentat.session_context import SESSION_CONTEXT
from mentat.vision.vision_manager import ScreenshotException


class ScreenshotCommand(Command, command_name="screenshot"):
    @override
    async def apply(self, *args: str) -> None:
        session_context = SESSION_CONTEXT.get()
        vision_manager = session_context.vision_manager
        stream = session_context.stream
        config = session_context.config
        conversation = session_context.conversation
        model = config.model

        if "gpt" in model:
            if "vision" not in model:
                stream.send(
                    "Using a version of gpt that doesn't support images. Changing to" " gpt-4-vision-preview",
                    style="warning",
                )
                config.model = "gpt-4-vision-preview"
        else:
            stream.send(
                "Can't determine if this model supports vision. Attempting anyway.",
                style="warning",
            )

        try:
            image = vision_manager.screenshot(*args)

            if len(args) == 0:
                path = "the current screen"
            else:
                path = args[0]
            conversation.add_user_message(f"A screenshot of {path}", image=image)
            stream.send(
                f"Screenshot taken for: {path}.",
                style="success",
            )
        except ScreenshotException:
            return  # Screenshot manager will print the error to stream.

    @override
    @classmethod
    def arguments(cls) -> List[CommandArgument]:
        return [CommandArgument("optional", ["path", "url"])]

    @override
    @classmethod
    def argument_autocompletions(cls, arguments: list[str], argument_position: int) -> list[str]:
        return get_command_filename_completions(arguments[-1])

    @override
    @classmethod
    def help_message(cls) -> str:
        return "Open a url or local file in a web browser and take a screenshot."



from __future__ import annotations

from pathlib import Path
from typing import Any

import attr

from mentat.errors import HistoryError, MentatError
from mentat.parsers.change_display_helper import (
    DisplayInformation,
    FileActionType,
    display_full_change,
)
from mentat.session_context import SESSION_CONTEXT
from mentat.session_input import ask_yes_no
from mentat.utils import get_relative_path


@attr.define(order=False)
class Replacement:
    """
    Represents that the 0-indexed lines from starting_line (inclusive) to ending_line (exclusive)
    should be replaced with new_lines
    """

    # Inclusive
    starting_line: int = attr.field()
    # Exclusive
    ending_line: int = attr.field()

    new_lines: list[str] = attr.field()

    def __lt__(self, other: Replacement):
        return self.ending_line < other.ending_line or (
            self.ending_line == other.ending_line and self.starting_line < other.ending_line
        )


async def _ask_user_change(
    text: str,
) -> bool:
    session_context = SESSION_CONTEXT.get()
    stream = session_context.stream

    stream.send(text, style="input")
    return await ask_yes_no(default_yes=True)


@attr.define
class FileEdit:
    """
    Represents that this file_path content should have specified Replacements applied to it.
    Can also represent that this file should be created, deleted, or is being renamed.
    """

    # Should be abs path
    file_path: Path = attr.field()
    replacements: list[Replacement] = attr.field(factory=list)
    is_creation: bool = attr.field(default=False)
    is_deletion: bool = attr.field(default=False)
    # Should be abs path
    rename_file_path: Path | None = attr.field(default=None)

    # Used for undo
    previous_file_lines: list[str] | None = attr.field(default=None)

    @file_path.validator  # pyright: ignore
    def is_abs_path(self, attribute: attr.Attribute[Path], value: Any):
        if not isinstance(value, Path):
            raise ValueError(f"File_path must be a Path, got {type(value)}")
        if not value.is_absolute():
            raise ValueError(f"File_path must be an absolute path, got {value}")

    def _display_creation(self, prefix: str = ""):
        added_lines = list[str]()
        for replacement in self.replacements:
            added_lines.extend(replacement.new_lines)
        display_information = DisplayInformation(self.file_path, [], added_lines, [], FileActionType.CreateFile)
        display_full_change(display_information, prefix=prefix)

    def _display_deletion(self, file_lines: list[str], prefix: str = ""):
        display_information = DisplayInformation(
            self.file_path,
            [],
            [],
            file_lines,
            FileActionType.DeleteFile,
        )
        display_full_change(display_information, prefix=prefix)

    def _display_rename(self, prefix: str = ""):
        display_information = DisplayInformation(
            self.file_path,
            [],
            [],
            [],
            FileActionType.RenameFile,
            new_name=self.rename_file_path,
        )
        display_full_change(display_information, prefix=prefix)

    def _display_replacement(self, replacement: Replacement, file_lines: list[str], prefix: str = ""):
        removed_block = file_lines[replacement.starting_line : replacement.ending_line]
        display_information = DisplayInformation(
            self.file_path,
            file_lines,
            replacement.new_lines,
            removed_block,
            FileActionType.UpdateFile,
            replacement.starting_line,
            replacement.ending_line,
            self.rename_file_path,
        )
        display_full_change(display_information, prefix=prefix)

    def _display_replacements(self, file_lines: list[str], prefix: str = ""):
        for replacement in self.replacements:
            self._display_replacement(replacement, file_lines, prefix=prefix)

    def display_full_edit(self, file_lines: list[str], prefix: str = ""):
        """Displays the full edit as if it were altering a file with the lines given"""
        if self.is_deletion:
            self._display_deletion(file_lines, prefix=prefix)
        if self.rename_file_path:
            self._display_rename(prefix=prefix)
        if self.is_creation:
            self._display_creation(prefix=prefix)
        else:
            self._display_replacements(file_lines, prefix=prefix)

    def is_valid(self) -> bool:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        code_context = session_context.code_context

        display_path = get_relative_path(self.file_path, session_context.cwd)

        if self.is_creation:
            if self.file_path.exists():
                stream.send(
                    f"File {display_path} already exists, canceling creation.",
                    style="warning",
                )
                return False
        else:
            if not self.file_path.exists():
                stream.send(
                    f"File {display_path} does not exist, canceling all edits to file.",
                    style="warning",
                )
                return False
            file_features_in_context = code_context.include_files.get(self.file_path, [])
            if not file_features_in_context or not all(
                any(f.interval.contains(i) for f in file_features_in_context)
                for r in self.replacements
                for i in range(r.starting_line + 1, r.ending_line + 1)
            ):
                stream.send(
                    f"File {display_path} not in context, canceling all edits to file.",
                    style="warning",
                )
                return False

        if self.rename_file_path is not None and self.rename_file_path.exists():
            rel_rename_path = None
            if self.rename_file_path.is_relative_to(session_context.cwd):
                rel_rename_path = self.rename_file_path.relative_to(session_context.cwd)
            stream.send(
                f"File {display_path} being renamed to existing file"
                f" {rel_rename_path or self.rename_file_path}, canceling rename.",
                style="warning",
            )
            self.rename_file_path = None
        return True

    async def filter_replacements(
        self,
    ) -> bool:
        session_context = SESSION_CONTEXT.get()
        code_file_manager = session_context.code_file_manager

        if self.is_creation:
            self._display_creation()
            if not await _ask_user_change("Create this file?"):
                return False
            file_lines = []
        else:
            file_lines = code_file_manager.file_lines[self.file_path].copy()

        if self.is_deletion:
            self._display_deletion(file_lines)
            if not await _ask_user_change("Delete this file?"):
                return False

        if self.rename_file_path is not None:
            self._display_rename()
            if not await _ask_user_change("Rename this file?"):
                self.rename_file_path = None

        if not self.is_creation:
            new_replacements = list[Replacement]()
            for replacement in sorted(self.replacements):
                self._display_replacement(replacement, file_lines)
                if await _ask_user_change("Keep this change?"):
                    new_replacements.append(replacement)
            self.replacements = new_replacements

        return self.is_creation or self.is_deletion or (self.rename_file_path is not None) or len(self.replacements) > 0

    def resolve_conflicts(self):
        self.replacements.sort(reverse=True)
        for index, replacement in enumerate(self.replacements):
            for other in self.replacements[index + 1 :]:
                if other.ending_line > replacement.starting_line and other.starting_line < replacement.ending_line:
                    # Overlap conflict
                    other.ending_line = replacement.starting_line
                    other.starting_line = min(other.starting_line, other.ending_line)
                elif (
                    other.ending_line == other.starting_line
                    and replacement.ending_line == replacement.starting_line
                    and replacement.starting_line == other.starting_line
                ):
                    # Insertion conflict (nothing to do)
                    pass

    def get_updated_file_lines(self, file_lines: list[str]):
        self.replacements.sort(reverse=True)
        earliest_line = None
        for replacement in self.replacements:
            if earliest_line is not None and replacement.ending_line > earliest_line:
                # This should never happen if resolve conflicts is called
                raise MentatError("Error: Line overlap in Replacements")
            if replacement.ending_line > len(file_lines):
                file_lines += [""] * (replacement.ending_line - len(file_lines))
            earliest_line = replacement.starting_line
            file_lines = (
                file_lines[: replacement.starting_line] + replacement.new_lines + file_lines[replacement.ending_line :]
            )
        return file_lines

    def undo(self):
        ctx = SESSION_CONTEXT.get()

        prefix = "UNDO: "

        if self.is_creation:
            if not self.file_path.exists():
                raise HistoryError(f"File {self.file_path} does not exist; unable to delete")
            ctx.code_file_manager.delete_file(self.file_path)

            self._display_creation(prefix=prefix)
            ctx.stream.send(f"Creation of file {self.file_path} undone", style="success")
            return

        if self.rename_file_path is not None:
            if self.file_path.exists():
                raise HistoryError(
                    f"File {self.file_path} already exists; unable to undo rename to" f" {self.rename_file_path}"
                )
            if not self.rename_file_path.exists():
                raise HistoryError(
                    f"File {self.rename_file_path} does not exist; unable to undo" f" rename from {self.file_path}"
                )
            ctx.code_file_manager.rename_file(self.rename_file_path, self.file_path)

            self._display_rename(prefix=prefix)
            ctx.stream.send(
                f"Rename of file {self.file_path} to {self.rename_file_path} undone",
                style="success",
            )

        if self.is_deletion:
            if self.file_path.exists():
                raise HistoryError(f"File {self.file_path} already exists; unable to re-create")
            if not self.previous_file_lines:
                # Should never happen
                raise ValueError("Previous file lines not set when undoing file deletion")
            ctx.code_file_manager.create_file(self.file_path, content="\n".join(self.previous_file_lines))

            self._display_deletion(self.previous_file_lines, prefix=prefix)
            ctx.stream.send(f"Deletion of file {self.file_path} undone", style="success")
        elif self.replacements:
            if not self.file_path.exists():
                raise HistoryError(f"File {self.file_path} does not exist; unable to undo edit")
            if not self.previous_file_lines:
                # Should never happen
                raise ValueError("Previous file lines not set when undoing file edit")

            ctx.code_file_manager.write_to_file(self.file_path, self.previous_file_lines)
            self._display_replacements(self.previous_file_lines, prefix=prefix)
            ctx.stream.send(f"Edits to file {self.file_path} undone", style="success")



import logging
import shlex
from typing import Any

from mentat.command.command import Command
from mentat.errors import SessionExit
from mentat.session_context import SESSION_CONTEXT
from mentat.session_stream import StreamMessage


async def _get_input_request(**kwargs: Any) -> StreamMessage:
    session_context = SESSION_CONTEXT.get()
    stream = session_context.stream

    message = stream.send("", channel="input_request", **kwargs)
    response = await stream.recv(f"input_request:{message.id}")
    logging.debug(f"User Input: {response.data}")
    return response


async def collect_user_input(command_autocomplete: bool = False) -> StreamMessage:
    """
    Listens for user input on a new channel

    send a message requesting user to send a response
    create a new broadcast channel that listens for the input
    close the channel after receiving the input
    """

    response = await _get_input_request(command_autocomplete=command_autocomplete)
    # Quit on q
    if isinstance(response.data, str) and response.data.strip() == "q":
        raise SessionExit

    return response


async def ask_yes_no(default_yes: bool) -> bool:
    session_context = SESSION_CONTEXT.get()
    stream = session_context.stream

    while True:
        # TODO: combine this into a single message (include content)
        stream.send("(Y/n)" if default_yes else "(y/N)")
        response = await collect_user_input()
        content = response.data.strip().lower()
        if content in ["y", "n", ""]:
            break
    return content == "y" or (content != "n" and default_yes)


async def collect_input_with_commands() -> StreamMessage:
    ctx = SESSION_CONTEXT.get()

    response = await collect_user_input(command_autocomplete=True)
    while isinstance(response.data, str) and response.data.startswith("/"):
        try:
            # We only use shlex to split the arguments, not the command itself
            arguments = shlex.split(" ".join(response.data.split(" ")[1:]))
            command = Command.create_command(response.data[1:].split(" ")[0])
            await command.apply(*arguments)
            await ctx.code_context.refresh_context_display()
        except ValueError as e:
            ctx.stream.send(f"Error processing command arguments: {e}", style="error")
        response = await collect_user_input(command_autocomplete=True)
    return response



from __future__ import annotations

import json
from typing import Optional

import attr

from mentat.transcripts import Transcript


@attr.define
class BenchmarkResult:
    name: str = attr.ib()
    family: Optional[str] = attr.ib(default=None)
    cost: Optional[float] = attr.ib(default=None, metadata={"aggregation": "sum"})
    tokens: Optional[int] = attr.ib(default=None, metadata={"aggregation": "average"})
    count: int = attr.ib(default=1, metadata={"aggregation": "sum"})
    iterations: Optional[int] = attr.ib(default=None, metadata={"aggregation": "histogram"})
    transcript: Optional[Transcript] = attr.ib(default=None, metadata={"display": "transcript"})
    instructions: Optional[str] = attr.ib(default=None, metadata={"display": "text"})
    code: Optional[str] = attr.ib(default=None, metadata={"display": "code"})
    test_output: Optional[str] = attr.ib(default=None, metadata={"formatted_name": "Test output", "display": "code"})
    run_error: Optional[str] = attr.ib(default=None, metadata={"formatted_name": "Run Error", "display": "code"})
    response: Optional[str] = attr.ib(default=None, metadata={"formatted_name": "Analysis", "display": "text"})
    reason: Optional[str] = attr.ib(default=None, metadata={"aggregation": "histogram"})
    # For exercism benchmarks
    passed: Optional[bool] = attr.ib(default=None, metadata={"aggregation": "percent"})
    # New optional fields for benchmark results
    diff_grade: Optional[dict] = attr.ib(default=None, metadata={"display": "json"})
    response_grade: Optional[dict] = attr.ib(default=None, metadata={"display": "json"})
    comparison_grade: Optional[dict] = attr.ib(default=None, metadata={"display": "json"})
    verify: Optional[bool] = attr.ib(default=None, metadata={"aggregation": "percent"})
    off_by_one: Optional[bool] = attr.ib(default=None, metadata={"aggregation": "percent"})
    indentation_error: Optional[bool] = attr.ib(default=None, metadata={"aggregation": "percent"})
    syntax_error: Optional[bool] = attr.ib(default=None, metadata={"aggregation": "percent"})
    missing_functionality: Optional[bool] = attr.ib(default=None, metadata={"aggregation": "percent"})
    extra_functionality: Optional[bool] = attr.ib(default=None, metadata={"aggregation": "percent"})
    referenced_format: Optional[bool] = attr.ib(default=None, metadata={"aggregation": "percent"})

    def display_color(self) -> str:
        if self.passed is None:
            if self.indentation_error or self.off_by_one or self.syntax_error:
                return "grey"
            if self.missing_functionality or self.extra_functionality or self.referenced_format:
                return "yellow"
            if self.verify is not None:
                if self.verify:
                    return "green"
                else:
                    return "red"
            return "green"
        elif self.passed:
            return "green"
        else:
            return "red"

    def to_dict(self):
        return attr.asdict(self)

    def to_json(self):
        return json.dumps(self.to_dict())

    @classmethod
    def from_dict(cls, d) -> BenchmarkResult:
        return cls(**d)

    @classmethod
    def load_json(cls, json_str):
        return cls.from_dict(json.loads(json_str))



import asyncio
from asyncio import Event
from asyncio.tasks import Task
from pathlib import Path
from typing import List

from mentat.config import Config
from mentat.errors import MentatError
from mentat.session import Session
from mentat.session_stream import StreamMessageSource


class PythonClient:
    """A client for interacting with Mentat in python."""

    def __init__(
        self,
        cwd: Path = Path.cwd(),
        paths: List[Path] = [],
        exclude_paths: List[Path] = [],
        ignore_paths: List[Path] = [],
        diff: str | None = None,
        pr_diff: str | None = None,
        config: Config = Config(),
    ):
        """
        Initializes the PythonClient with configuration for interacting with Mentat.

        You must call `startup` to begin the client and `shutdown` to end it.

        Parameters:
        - cwd (Path): The current working directory for the client.
        - paths (List[Path]): A list of paths to include in the analysis.
        - exclude_paths (List[Path]): A list of paths to exclude from the analysis.
        - ignore_paths (List[Path]): A list of paths to ignore for the purpose of analysis.
        - diff (str | None): A treeish diff that mentat will be aware of when making changes.
        - pr_diff (str | None): Like diff but it compares to the common ancester.
        - config (Config): Configuration settings for the client.
        """
        self.cwd = cwd.expanduser().resolve()
        self.paths = paths
        self.exclude_paths = exclude_paths
        self.ignore_paths = ignore_paths
        self.diff = diff
        self.pr_diff = pr_diff
        self.config = config

        self._accumulated_message = ""
        self.stopped = Event()
        self._call_mentat_task = None

    async def _call_mentat(self, message: str):
        input_request_message = await self.session.stream.recv("input_request")
        self.session.stream.send(
            message,
            source=StreamMessageSource.CLIENT,
            channel=f"input_request:{input_request_message.id}",
        )
        if message.strip().lower() == "y":
            await self.wait_for_edit_completion()

        temp = self._accumulated_message
        self._accumulated_message = ""
        return temp

    async def call_mentat(self, message: str):
        """Call Mentat with a message and return the response.

        Behaves the same as talking to mentat as an application so you can use commands.
        """
        self._call_mentat_task = asyncio.create_task(self._call_mentat(message))
        try:
            return await self._call_mentat_task
        except asyncio.CancelledError:
            print(self.session.error)
            raise MentatError("Session failed")

    async def call_mentat_auto_accept(self, message: str) -> str:
        """Call Mentat with a message and then accept the edits."""
        response = await self.call_mentat(message)
        await self.call_mentat("y")
        return response

    async def wait_for_edit_completion(self):
        await self.session.stream.recv(channel="edits_complete")

    async def _accumulate_messages(self):
        async for message in self.session.stream.listen():
            end = "\n"
            if isinstance(message.extra.get("end"), str):
                end = message.extra["end"]
            self._accumulated_message += message.data + end

    async def _listen_for_client_exit(self):
        await self.session.stream.recv("client_exit")
        await self._stop()

    async def _listen_for_session_stopped(self):
        await self.session.stream.recv("session_stopped")
        await self._stop()

    async def startup(self):
        """Starts up the client, establishing a session and beginning to listen for messages and exit signals."""
        self.session = Session(
            self.cwd,
            self.paths,
            self.exclude_paths,
            self.ignore_paths,
            self.diff,
            self.pr_diff,
            self.config,
        )
        self.session.start()
        self.acc_task = asyncio.create_task(self._accumulate_messages())
        self.client_exit_task: Task[None] = asyncio.create_task(self._listen_for_client_exit())
        self.session_stopped_task: Task[None] = asyncio.create_task(self._listen_for_session_stopped())

    async def shutdown(self):
        """Initiates shutdown of the client, ensuring all tasks are cancelled and the session is properly closed.
        Sends the stop signal to the session and returns when client is fully shutdown.
        """
        self.session.stream.send(None, channel="session_exit")
        await self.stopped.wait()

    async def _stop(self):
        self.acc_task.cancel()
        self.client_exit_task.cancel()
        self.session_stopped_task.cancel()
        if self._call_mentat_task:
            self._call_mentat_task.cancel()
        self.stopped.set()

    def get_conversation(self):
        """Returns the current conversation context from the session."""
        return self.session.ctx.conversation

    def get_cost_tracker(self):
        """Returns the cost tracker from the session context."""
        return self.session.ctx.cost_tracker



import os
import subprocess

from benchmarks.exercise_runners.abstract_exercise_runner import AbstractExerciseRunner


class JavascriptExerciseRunner(AbstractExerciseRunner):
    def __init__(self, exercise):
        super().__init__(exercise, "js")
        if not os.path.exists("node_modules"):
            subprocess.run(["npm", "install"], stdout=subprocess.PIPE)
        self.test_file = self.dir / f"{exercise}.spec.js"
        with open(self.test_file, "r") as f:
            test_contents = f.read()
        with open(self.test_file, "w") as f:
            test_contents = f.write(test_contents.replace("xtest", "test"))

    def run_test(self):
        self._run_test_command(["./node_modules/jest/bin/jest.js", self.dir])

    def passed(self):
        try:
            with open(self.test_output_file, "r") as f:
                lines = f.readlines()
                return "FAIL" not in lines[0] and "PASS" in lines[0]
        except FileNotFoundError:
            return False



import json
import os
from pathlib import Path
from typing import Optional, Tuple

import attr

from benchmarks.benchmark_result import BenchmarkResult


class BenchmarkRunSummary:
    def __init__(
        self,
        summary: dict[str, Tuple[int | float | str, float]],
        metadata: Optional[dict] = None,
    ):
        self.summary = summary
        self.metadata = metadata
        self.display_string = self.display_string()

    def formatted_summary(self) -> dict[str, str]:
        formatted = {}
        total = self.summary["count"][0]
        for field in attr.fields(BenchmarkResult):
            if "aggregation" in field.metadata:
                name = field.name
                value, total_set = self.summary[name]
                if total_set == 0:
                    continue
                percent_set_display = ""
                if total_set < total:
                    percent_set_display = f" ({total_set}/{total})"
                formatted_value = ""
                aggregation_type = field.metadata["aggregation"]
                if aggregation_type == "average":
                    formatted_name = f"{name} (avg)"
                else:
                    formatted_name = name

                if isinstance(value, float):
                    formatted_value = f"{value:.2f}"
                elif isinstance(value, dict):
                    formatted_value = ", ".join(f"{k}: {v}" for k, v in value.items())
                else:
                    formatted_value = str(value)

                # Add units based on aggregation type
                if aggregation_type == "sum" and "cost" in formatted_name:
                    formatted[formatted_name] = f"${formatted_value} {percent_set_display}"
                elif aggregation_type == "percent":
                    formatted[formatted_name] = f"{formatted_value}% {percent_set_display}"
                else:
                    formatted[formatted_name] = f"{formatted_value} {percent_set_display}"

        return formatted

    def display_string(self) -> str:
        return ", ".join(f"{name}: {value}" for name, value in self.formatted_summary().items())

    def to_json(self) -> str:
        return json.dumps(
            {
                "summary": self.summary,
                "metadata": self.metadata,
            },
            indent=4,
        )

    @classmethod
    def load_json(cls, json_str: str) -> "BenchmarkRunSummary":
        data = json.loads(json_str)
        return cls(data["summary"], data["metadata"])

    @classmethod
    def load_file(cls, file: Path) -> "BenchmarkRunSummary":
        file_name = os.path.basename(file)
        with open(file, "r") as f:
            summary = cls.load_json(f.read())
        summary.metadata["file"] = file_name  # Used for links
        return summary



import glob
import json
import re
from typing import TypedDict

from openai.types.chat import ChatCompletionContentPartParam, ChatCompletionMessageParam

from mentat.logging_config import logs_path


class UserMessage(TypedDict):
    message: list[ChatCompletionContentPartParam] | str
    # We need this field so that it is included when we convert to JSON
    prior_messages: None


class ModelMessage(TypedDict, total=False):
    message: str
    prior_messages: list[ChatCompletionMessageParam]

    # Used to mark different types of messages; e.g., an agent message that isn't part of the regular conversation
    # NotRequired isn't available until 3.11, so we have to use total=False instead
    message_type: str


TranscriptMessage = UserMessage | ModelMessage


class Transcript(TypedDict):
    id: str
    messages: list[TranscriptMessage]


def get_transcript_logs() -> list[Transcript]:
    transcripts = glob.glob(str(logs_path / "transcript_*"))
    ans = list[Transcript]()
    for transcript_path in transcripts:
        match = re.search(r"transcript_(.+).log", transcript_path)
        if match is None:
            continue
        timestamp = match.group(1)

        with open(transcript_path, "r") as f:
            transcript = f.readlines()
        if len(transcript) == 0:
            continue

        transcript = json.loads("[" + ", ".join(transcript) + "]")
        ans.append(Transcript(id=timestamp, messages=transcript))

    return sorted(ans, key=lambda x: x["id"], reverse=True)



import re
from typing import Optional

import packaging.version
import requests

from mentat.session_context import SESSION_CONTEXT
from mentat.utils import mentat_dir_path
from mentat.version import __version__


def get_changelog() -> Optional[str]:
    try:
        response = requests.get("https://raw.githubusercontent.com/AbanteAI/mentat/main/CHANGELOG.rst")
        if response.status_code == 200:
            return response.text
        else:
            return None
    except Exception:
        return None


def get_latest_changelog(full_changelog: Optional[str] = None) -> Optional[str]:
    if full_changelog is None:
        full_changelog = get_changelog()
    if full_changelog is None:
        return None
    try:
        sections = re.split("\n[^\n]+\n-{1,}\n", full_changelog)
        return sections[1].strip()
    except Exception:
        return None


def check_version():
    ctx = SESSION_CONTEXT.get()

    try:
        response = requests.get("https://pypi.org/pypi/mentat/json")
        data = response.json()
        latest_version = data["info"]["version"]

        if packaging.version.parse(__version__) < packaging.version.parse(latest_version):
            ctx.stream.send(
                f"Version v{latest_version} of Mentat is available. If pip was used to"
                " install Mentat, upgrade with:",
                style="warning",
            )
            ctx.stream.send("pip install --upgrade mentat", style="warning")
            changelog = get_latest_changelog()
            if changelog:
                ctx.stream.send("Upgrade for the following features/improvements:", style="warning")
                ctx.stream.send(changelog, style="warning")

        else:
            last_version_check_file = mentat_dir_path / "last_version_check"
            if last_version_check_file.exists():
                with open(last_version_check_file, "r") as f:
                    last_version_check = f.read()
                if packaging.version.parse(last_version_check) < packaging.version.parse(__version__):
                    changelog = get_latest_changelog()
                    if changelog:
                        ctx.stream.send(f"Thanks for upgrading to v{__version__}.", style="info")
                        ctx.stream.send("Changes in this version:", style="info")
                        ctx.stream.send(changelog, style="info")
            with open(last_version_check_file, "w") as f:
                f.write(__version__)
    except Exception as err:
        ctx.stream.send(f"Error checking for most recent version: {err}", style="error")


def check_model():
    ctx = SESSION_CONTEXT.get()
    model = ctx.config.model
    if "gpt-4" not in model:
        ctx.stream.send(
            "Warning: Mentat has only been tested on GPT-4. You may experience"
            " issues with quality. This model may not be able to respond in"
            " mentat's edit format.",
            style="warning",
        )
        if "gpt-3.5" not in model:
            ctx.stream.send(
                "Warning: Mentat does not know how to calculate costs or context" " size for this model.",
                style="warning",
            )



from typing import List

from typing_extensions import override

from mentat.command.command import Command, CommandArgument
from mentat.session_context import SESSION_CONTEXT

help_message_width = 60


def _arg_message(arg: CommandArgument) -> str:
    if isinstance(arg.description, str):
        cur_message = arg.description
    else:
        cur_message = "|".join(arg.description)

    match arg.arg_type:
        case "required":
            cur_message = "<" + cur_message + ">"
        case "optional":
            cur_message = "[" + cur_message + "]"
        case "literal":
            pass

    if arg.repeatable:
        cur_message += " ..."

    return cur_message


class HelpCommand(Command, command_name="help"):
    @override
    async def apply(self, *args: str) -> None:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream

        if not args:
            commands = Command.get_command_names()
        else:
            commands = args
        for command_name in commands:
            if command_name not in Command._registered_commands:
                stream.send(f"Error: Command {command_name} does not exist.", style="error")
            else:
                command_cls = Command._registered_commands[command_name]
                arguments = command_cls.arguments()
                argument_message = [_arg_message(arg) for arg in arguments]

                help_message = command_cls.help_message()
                message = " ".join([f"/{command_name}"] + argument_message).ljust(help_message_width) + help_message
                stream.send(message)

    @override
    @classmethod
    def arguments(cls) -> List[CommandArgument]:
        return [CommandArgument("optional", "command", repeatable=True)]

    @override
    @classmethod
    def argument_autocompletions(cls, arguments: list[str], argument_position: int) -> list[str]:
        return Command.get_command_names()

    @override
    @classmethod
    def help_message(cls) -> str:
        return "Show information on available commands."



from benchmarks.exercise_runners.clojure_exercise_runner import ClojureExerciseRunner
from benchmarks.exercise_runners.javascript_exercise_runner import (
    JavascriptExerciseRunner,
)
from benchmarks.exercise_runners.python_exercise_runner import PythonExerciseRunner


class ExerciseRunnerFactory:
    RUNNERS = {
        "clojure": ClojureExerciseRunner,
        "javascript": JavascriptExerciseRunner,
        "python": PythonExerciseRunner,
    }

    @classmethod
    def create(cls, language, exercise):
        return cls.RUNNERS[language](exercise)



import difflib
from pathlib import Path
from typing import List

from openai.types.chat import (
    ChatCompletionAssistantMessageParam,
    ChatCompletionMessageParam,
    ChatCompletionSystemMessageParam,
)

from mentat.errors import MentatError
from mentat.llm_api_handler import prompt_tokens
from mentat.parsers.change_display_helper import get_lexer, highlight_text
from mentat.parsers.file_edit import FileEdit
from mentat.parsers.git_parser import GitParser
from mentat.parsers.streaming_printer import FormattedString, send_formatted_string
from mentat.prompts.prompts import read_prompt
from mentat.session_context import SESSION_CONTEXT
from mentat.transcripts import ModelMessage
from mentat.utils import get_relative_path

revisor_prompt_filename = Path("revisor_prompt.txt")
revisor_prompt = read_prompt(revisor_prompt_filename)


def _get_stored_lines(file_edit: FileEdit) -> List[str]:
    ctx = SESSION_CONTEXT.get()

    if file_edit.is_creation:
        return []
    else:
        return ctx.code_file_manager.file_lines[file_edit.file_path].copy()


def _file_edit_diff(file_edit: FileEdit) -> str:
    stored_lines = _get_stored_lines(file_edit)
    new_lines = file_edit.get_updated_file_lines(stored_lines)
    diff = list(difflib.unified_diff(stored_lines, new_lines, lineterm=""))
    return "\n".join(diff)


async def revise_edit(file_edit: FileEdit):
    ctx = SESSION_CONTEXT.get()

    # No point in revising deletion edits
    if file_edit.is_deletion:
        return
    diff = _file_edit_diff(file_edit)
    # There should always be a user_message by the time we're revising
    user_message = list(
        filter(
            lambda message: message["role"] == "user",
            await ctx.conversation.get_messages(),
        )
    )[-1]
    user_message["content"] = f"User Request:\n{user_message.get('content')}"
    messages: List[ChatCompletionMessageParam] = [
        ChatCompletionSystemMessageParam(content=revisor_prompt, role="system"),
        user_message,
        ChatCompletionSystemMessageParam(content=f"Diff:\n{diff}", role="system"),
    ]
    code_message = await ctx.code_context.get_code_message(prompt_tokens(messages, ctx.config.model))
    messages.insert(1, ChatCompletionSystemMessageParam(content=code_message, role="system"))

    ctx.stream.send(
        "\nRevising edits for file" f" {get_relative_path(file_edit.file_path, ctx.cwd)}...",
        style="info",
    )
    response = await ctx.llm_api_handler.call_llm_api(messages, model=ctx.config.model, stream=False)
    message = response.choices[0].message.content or ""
    messages.append(ChatCompletionAssistantMessageParam(content=message, role="assistant"))
    ctx.conversation.add_transcript_message(
        ModelMessage(message=message, prior_messages=messages, message_type="revisor")
    )

    # Sometimes the model wraps response in a ```diff ``` block
    # I believe new prompt fixes this but this makes sure it never interferes
    if message.startswith("```diff\n"):
        message = message[8:]
        if message.endswith("\n```"):
            message = message[:-4]

    # This makes it more similar to a git diff so that we can use the pre existing git diff parser
    message = "\n".join(message.split("\n")[2:])  # remove leading +++ and ---
    post_diff = "diff --git a/file b/file\nindex 0000000..0000000\n--- a/file\n+++" f" b/file\n{message}"
    parsed_response = GitParser().parse_llm_response(post_diff)

    # Only modify the replacements of the current file edit
    # (the new file edit doesn't know about file creation or renaming)
    # Additionally, since we do this one at a time there should only ever be 1 file edit.
    if parsed_response.file_edits:
        stored_lines = _get_stored_lines(file_edit)
        pre_lines = file_edit.get_updated_file_lines(stored_lines)
        file_edit.replacements = parsed_response.file_edits[0].replacements
        post_lines = file_edit.get_updated_file_lines(stored_lines)

        diff_lines = difflib.unified_diff(pre_lines, post_lines, lineterm="")
        diff_diff: List[FormattedString] = []
        lexer = get_lexer(file_edit.file_path)
        for line in diff_lines:
            if line.startswith("---"):
                diff_diff.append(f"{line}{file_edit.file_path}")
            elif line.startswith("+++"):
                new_name = file_edit.rename_file_path if file_edit.rename_file_path is not None else file_edit.file_path
                diff_diff.append(f"{line}{new_name}")
            elif line.startswith("@@"):
                diff_diff.append(line)
            elif line.startswith("+"):
                diff_diff.append((line, {"color": "green"}))
            elif line.startswith("-"):
                diff_diff.append((line, {"color": "red"}))
            elif line.startswith(" "):
                diff_diff.append(highlight_text(line, lexer))
            else:
                raise MentatError("Invalid Diff")
        if diff_diff:
            ctx.stream.send("Revision diff:", style="info")
            ctx.stream.send("", delimiter=True)
            for line in diff_diff:
                send_formatted_string(line)
            ctx.stream.send("", delimiter=True)
        ctx.cost_tracker.display_last_api_call()


async def revise_edits(file_edits: List[FileEdit]):
    for file_edit in file_edits:
        # We could do all edits asynchronously; of course, this risks getting rate limited and probably not worth effort
        await revise_edit(file_edit)



from __future__ import annotations

import math
import re
from pathlib import Path

import attr


def split_intervals_from_path(custom_path: str | Path) -> tuple[Path, str]:
    match = re.match(
        r"(.*?):((\d+-\d+|\d+)(,\d+(-\d+)?)*$)",  # One or more intervals/numbers
        str(custom_path),
    )
    if match:
        path, intervals = match.groups()[0], match.groups()[1]
        return Path(path), intervals
    else:
        return Path(custom_path), ""


def parse_intervals(interval_string: str) -> list[Interval]:
    try:
        intervals = list[Interval]()
        for interval in interval_string.split(","):
            interval = interval.split("-", 1)
            if len(interval) == 1:
                interval = Interval(int(interval[0]), int(interval[0]) + 1)
            else:
                interval = Interval(int(interval[0]), int(interval[1]))
            if interval.end <= interval.start:
                continue
            intervals.append(interval)
        return intervals
    except (ValueError, IndexError):
        return []


# Unfortunately there is no any way to set class properties with attrs so we can't make this part of Interval
INTERVAL_FILE_END = math.inf


@attr.define(order=True, frozen=True)
class Interval:
    """
    1-indexed interval of file lines, inclusive start, exclusive end
    """

    start: int | float = attr.field()
    end: int | float = attr.field()

    def contains(self, line_number: int) -> bool:
        return self.start <= line_number < self.end

    def intersects(self, other: Interval) -> bool:
        return self.start < other.end and other.start < self.end

    def whole_file(self) -> bool:
        return self.start == 1 and self.end == INTERVAL_FILE_END

    def __str__(self) -> str:
        if self.end == self.start + 1:
            return f"{self.start}"
        else:
            return f"{self.start}-{self.end}"

    @staticmethod
    def from_string(interval_string: str) -> Interval:
        try:
            interval_parts = interval_string.split("-")
        except ValueError:
            return Interval(1, INTERVAL_FILE_END)

        if len(interval_parts) != 2:
            # corrupt interval string, make it whole file
            return Interval(1, INTERVAL_FILE_END)

        start, end = interval_parts
        return Interval(int(start), int(end))



from abc import ABC, abstractmethod

from mentat.code_feature import CodeFeature


class FeatureFilter(ABC):
    """
    Tools to pare down a list of Features to a final list to be put in an LLMs context.
    Despite the name they may not be pure filters:
    * New CodeFeatures may be introduced by splitting Features into intervals.
    * The CodeFeatures may be reordered by priority in future steps.

    A feature filter may want more information than the list of features and user prompt. If so it
    can get it from the config or have the information passed into its constructor.
    """

    @abstractmethod
    async def filter(
        self,
        features: list[CodeFeature],
    ) -> list[CodeFeature]:
        raise NotImplementedError()



from __future__ import annotations

import os
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Set, TypedDict, Union

from mentat.code_feature import (
    CodeFeature,
    get_code_message_from_features,
    get_consolidated_feature_refs,
    split_file_into_intervals,
)
from mentat.diff_context import DiffContext
from mentat.errors import PathValidationError
from mentat.feature_filters.default_filter import DefaultFilter
from mentat.feature_filters.embedding_similarity_filter import EmbeddingSimilarityFilter
from mentat.include_files import (
    PathType,
    get_code_features_for_path,
    get_path_type,
    get_paths_for_directory,
    is_file_text_encoded,
    match_path_with_patterns,
    validate_and_format_path,
)
from mentat.interval import parse_intervals, split_intervals_from_path
from mentat.llm_api_handler import count_tokens, get_max_tokens
from mentat.session_context import SESSION_CONTEXT
from mentat.session_stream import SessionStream


class ContextStreamMessage(TypedDict):
    cwd: str
    diff_context_display: Optional[str]
    auto_context_tokens: int
    features: List[str]
    git_diff_paths: List[str]
    git_untracked_paths: List[str]
    total_tokens: int
    maximum_tokens: int
    total_cost: float


class CodeContext:
    def __init__(
        self,
        stream: SessionStream,
        cwd: Path,
        diff: Optional[str] = None,
        pr_diff: Optional[str] = None,
        ignore_patterns: Iterable[Path | str] = [],
    ):
        self.diff = diff
        self.pr_diff = pr_diff
        self.ignore_patterns = set(Path(p) for p in ignore_patterns)

        self.diff_context = DiffContext(stream, cwd, self.diff, self.pr_diff)

        self.include_files: Dict[Path, List[CodeFeature]] = {}
        self.ignore_files: Set[Path] = set()

    async def refresh_context_display(self):
        """
        Sends a message to the client with the code context. It is called in the main loop.
        """
        ctx = SESSION_CONTEXT.get()

        diff_context_display = self.diff_context.get_display_context()

        features = get_consolidated_feature_refs(
            [feature for file_features in self.include_files.values() for feature in file_features]
        )
        git_diff_paths = [str(p) for p in self.diff_context.diff_files()]
        git_untracked_paths = [str(p) for p in self.diff_context.untracked_files()]

        total_tokens = await ctx.conversation.count_tokens(include_code_message=True)

        total_cost = ctx.cost_tracker.total_cost

        data = ContextStreamMessage(
            cwd=str(ctx.cwd),
            diff_context_display=diff_context_display,
            auto_context_tokens=ctx.config.auto_context_tokens,
            features=features,
            git_diff_paths=git_diff_paths,
            git_untracked_paths=git_untracked_paths,
            total_tokens=total_tokens,
            maximum_tokens=get_max_tokens(),
            total_cost=total_cost,
        )
        ctx.stream.send(data, channel="context_update")

    async def get_code_message(
        self,
        prompt_tokens: int,
        prompt: Optional[str] = None,
        expected_edits: Optional[list[str]] = None,  # for training/benchmarking
    ) -> str:
        """
        Retrieves the current code message.
        'prompt' argument is embedded and used to search for similar files when auto-context is enabled.
        If prompt is empty, auto context won't be used.
        'prompt_tokens' argument is the total number of tokens used by the prompt before the code message,
        used to ensure that the code message won't overflow the model's context size
        """
        session_context = SESSION_CONTEXT.get()
        config = session_context.config
        model = config.model

        # Setup code message metadata
        code_message = list[str]()

        # Since there is no way of knowing when the git diff changes,
        # we just refresh the cache every time get_code_message is called
        self.diff_context.refresh()
        if self.diff_context.diff_files():
            code_message += [
                "Diff References:",
                f' "-" = {self.diff_context.name}',
                ' "+" = Active Changes',
                "",
            ]

        code_message += ["Code Files:\n"]

        # Get auto included features
        if config.auto_context_tokens > 0 and prompt:
            meta_tokens = count_tokens("\n".join(code_message), model, full_message=True)

            # Calculate user included features token size
            include_files_message = get_code_message_from_features(
                [feature for file_features in self.include_files.values() for feature in file_features]
            )
            include_files_tokens = count_tokens("\n".join(include_files_message), model, full_message=False)

            tokens_used = prompt_tokens + meta_tokens + include_files_tokens
            auto_tokens = min(
                get_max_tokens() - tokens_used - config.token_buffer,
                config.auto_context_tokens,
            )
            features = self.get_all_features()
            feature_filter = DefaultFilter(auto_tokens, prompt, expected_edits)
            self.include_features(await feature_filter.filter(features))

            # TODO: We want to show the auto included features immediately, but refreshing the context display
            # also refreshes the token count per message, which calls this function again causing an infinite loop.
            # To fix this, we should completely separate the token count per message from the context display message
            # await self.refresh_context_display()

        include_features = [feature for file_features in self.include_files.values() for feature in file_features]
        code_message += get_code_message_from_features(include_features)

        return "\n".join(code_message)

    def get_all_features(
        self,
        max_chars: int = 100000,
        split_intervals: bool = True,
    ) -> list[CodeFeature]:
        """
        Retrieves every CodeFeature under the cwd. If files_only is True the features won't be split into intervals
        """
        session_context = SESSION_CONTEXT.get()

        abs_exclude_patterns: Set[Path] = set()
        for pattern in self.ignore_patterns.union(session_context.config.file_exclude_glob_list):
            if not Path(pattern).is_absolute():
                abs_exclude_patterns.add(session_context.cwd / pattern)
            else:
                abs_exclude_patterns.add(Path(pattern))

        all_features: List[CodeFeature] = []
        for path in get_paths_for_directory(path=session_context.cwd, exclude_patterns=abs_exclude_patterns):
            if not is_file_text_encoded(path) or os.path.getsize(path) > max_chars:
                continue

            if not split_intervals:
                _feature = CodeFeature(path)
                all_features.append(_feature)
            else:
                full_feature = CodeFeature(path)
                _split_features = split_file_into_intervals(full_feature)
                all_features += _split_features

        return sorted(all_features, key=lambda f: f.path)

    def include_features(self, code_features: Iterable[CodeFeature]):
        """
        Adds the given code features to context. If the feature is already included, it will not be added.
        """
        included_paths: Set[Path] = set()
        for code_feature in code_features:
            if code_feature.path not in self.include_files:
                self.include_files[code_feature.path] = [code_feature]
                included_paths.add(Path(str(code_feature)))
            else:
                code_feature_not_included = True
                for included_code_feature in self.include_files[code_feature.path]:
                    # Intervals can still overlap if user includes intervals different than what ctags breaks up,
                    # but we merge when making code message and don't duplicate lines
                    if (
                        included_code_feature.interval == code_feature.interval
                        # No need to include an interval if the entire file is already included
                        or included_code_feature.interval.whole_file()
                    ):
                        code_feature_not_included = False
                        break
                if code_feature_not_included:
                    if code_feature.interval.whole_file():
                        self.include_files[code_feature.path] = []
                    self.include_files[code_feature.path].append(code_feature)
                    included_paths.add(Path(str(code_feature)))
        return included_paths

    def include(self, path: Path | str, exclude_patterns: Iterable[Path | str] = []) -> Set[Path]:
        """
        Add paths to the context

        Paths that are already in the context and invalid paths are ignored.

        The following are behaviors for each `path` type:

        - File: the file is added to the context.
        - File interval: the file is added to the context, and the interval is added to the file
        - Directory: all files in the directory are recursively added to the context
        - Glob pattern: all files that match the glob pattern are added to the context

        Args:
            `path`: can be a relative or absolute file path, file interval path, directory, or glob pattern.

        Return:
            A set of paths that have been successfully included in the context
        """
        session_context = SESSION_CONTEXT.get()

        path = Path(path)

        abs_exclude_patterns: Set[Path] = set()
        all_exclude_patterns: Set[Union[str, Path]] = set(
            [
                *exclude_patterns,
                *self.ignore_patterns,
                *session_context.config.file_exclude_glob_list,
            ]
        )
        for pattern in all_exclude_patterns:
            if not Path(pattern).is_absolute():
                abs_exclude_patterns.add(session_context.cwd / pattern)
            else:
                abs_exclude_patterns.add(Path(pattern))

        try:
            code_features = get_code_features_for_path(
                path=path,
                cwd=session_context.cwd,
                exclude_patterns=abs_exclude_patterns,
            )
        except PathValidationError as e:
            session_context.stream.send(str(e), style="error")
            return set()

        return self.include_features(code_features)

    def _exclude_file(self, path: Path) -> Path | None:
        session_context = SESSION_CONTEXT.get()
        if path in self.include_files:
            del self.include_files[path]
            return path
        else:
            session_context.stream.send(f"Path {path} not in context", style="error")

    def _exclude_file_interval(self, path: Path) -> Set[Path]:
        session_context = SESSION_CONTEXT.get()

        excluded_paths: Set[Path] = set()

        interval_path, interval_str = split_intervals_from_path(path)
        if interval_path not in self.include_files:
            session_context.stream.send(f"Path {interval_path} not in context", style="error")
            return excluded_paths

        intervals = parse_intervals(interval_str)
        included_code_features: List[CodeFeature] = []
        for code_feature in self.include_files[interval_path]:
            if code_feature.interval not in intervals:
                included_code_features.append(code_feature)
            else:
                excluded_paths.add(Path(str(code_feature)))

        if len(included_code_features) == 0:
            del self.include_files[interval_path]
        else:
            self.include_files[interval_path] = included_code_features

        return excluded_paths

    def _exclude_directory(self, path: Path) -> Set[Path]:
        excluded_paths: Set[Path] = set()

        paths_to_exclude: Set[Path] = set()
        for included_path in self.include_files:
            if path in included_path.parents:
                paths_to_exclude.add(included_path)
        for excluded_path in paths_to_exclude:
            del self.include_files[excluded_path]
            excluded_paths.add(excluded_path)

        return excluded_paths

    def _exclude_glob(self, path: Path) -> Set[Path]:
        excluded_paths: Set[Path] = set()

        paths_to_exclude: Set[Path] = set()
        for included_path in self.include_files:
            if match_path_with_patterns(included_path, set([path])):
                paths_to_exclude.add(included_path)
        for excluded_path in paths_to_exclude:
            del self.include_files[excluded_path]
            excluded_paths.add(excluded_path)

        return excluded_paths

    def exclude(self, path: Path | str) -> Set[Path]:
        """Remove code from the context

        Paths that are not in the context and invalid paths are ignored.

        The following are behaviors for each `path` type:

        - File: the file is removed from the context
        - File interval: the interval is removed from the file. If the file doesn't have the exact interval specified,
          nothing is removed.
        - Directory: all files in the directory are removed from the context
        - Glob pattern: all files that match the glob pattern are removed from the context

        Args:
            `path`: can be a relative or absolute file path, file interval path, directory, or glob pattern.

        Return:
            A set of paths that have been successfully excluded from the context
        """
        session_context = SESSION_CONTEXT.get()

        path = Path(path)
        excluded_paths: Set[Path] = set()
        try:
            validated_path = validate_and_format_path(path, session_context.cwd, check_for_text=False)
            match get_path_type(validated_path):
                case PathType.FILE:
                    excluded_path = self._exclude_file(validated_path)
                    if excluded_path:
                        excluded_paths.add(excluded_path)
                case PathType.FILE_INTERVAL:
                    excluded_paths.update(self._exclude_file_interval(validated_path))
                case PathType.DIRECTORY:
                    excluded_paths.update(self._exclude_directory(validated_path))
                case PathType.GLOB:
                    excluded_paths.update(self._exclude_glob(validated_path))
        except PathValidationError as e:
            session_context.stream.send(str(e), style="error")
            pass

        return excluded_paths

    async def search(
        self,
        query: str,
        max_results: int | None = None,
    ) -> list[tuple[CodeFeature, float]]:
        """Return the top n features that are most similar to the query."""

        all_features = self.get_all_features()

        embedding_similarity_filter = EmbeddingSimilarityFilter(query)
        all_features_sorted = await embedding_similarity_filter.score(all_features)
        if max_results is None:
            return all_features_sorted
        else:
            return all_features_sorted[:max_results]

    def to_simple_context_dict(self) -> dict[str, list[str]]:
        """Return a simple dictionary representation of the code context"""

        simple_dict: dict[str, list[str]] = {}
        for path, features in self.include_files.items():
            simple_dict[str(path.absolute())] = [str(feature) for feature in features]
        return simple_dict

    def from_simple_context_dict(self, simple_dict: dict[str, list[str]]):
        """Load the code context from a simple dictionary representation"""

        for path_str, features_str in simple_dict.items():
            path = Path(path_str)
            features_for_path: List[CodeFeature] = []

            for feature_str in features_str:
                feature_path = Path(feature_str)

                # feature_path is already absolute, so cwd doesn't matter
                current_features = get_code_features_for_path(feature_path, cwd=Path("/"))
                features_for_path += list(current_features)

            self.include_files[path] = features_for_path



from typing import List

from typing_extensions import override

from mentat.command.command import Command, CommandArgument
from mentat.session_context import SESSION_CONTEXT


class ClearCommand(Command, command_name="clear"):
    @override
    async def apply(self, *args: str) -> None:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        conversation = session_context.conversation

        conversation.clear_messages()
        message = "Message history cleared."
        stream.send(message, style="success")

    @override
    @classmethod
    def arguments(cls) -> List[CommandArgument]:
        return []

    @override
    @classmethod
    def argument_autocompletions(cls, arguments: list[str], argument_position: int) -> list[str]:
        return []

    @override
    @classmethod
    def help_message(cls) -> str:
        return "Clear the current message history."



import asyncio
import logging
import os
import traceback
from asyncio import CancelledError, Event, Task
from pathlib import Path
from typing import Any, Coroutine, List, Optional, Set
from uuid import uuid4

import attr
import sentry_sdk
from openai import (
    APITimeoutError,
    BadRequestError,
    PermissionDeniedError,
    RateLimitError,
)

from mentat.agent_handler import AgentHandler
from mentat.auto_completer import AutoCompleter
from mentat.code_context import CodeContext
from mentat.code_edit_feedback import get_user_feedback_on_edits
from mentat.code_file_manager import CodeFileManager
from mentat.config import Config
from mentat.conversation import Conversation
from mentat.cost_tracker import CostTracker
from mentat.ctags import ensure_ctags_installed
from mentat.errors import MentatError, ReturnToUser, SessionExit, UserError
from mentat.llm_api_handler import LlmApiHandler, is_test_environment
from mentat.logging_config import setup_logging
from mentat.parsers.file_edit import FileEdit
from mentat.revisor.revisor import revise_edits
from mentat.sampler.sampler import Sampler
from mentat.sentry import sentry_init
from mentat.session_context import SESSION_CONTEXT, SessionContext
from mentat.session_input import collect_input_with_commands
from mentat.session_stream import SessionStream
from mentat.splash_messages import check_model, check_version
from mentat.utils import mentat_dir_path
from mentat.vision.vision_manager import VisionManager


class Session:
    """
    The server for Mentat.
    To stop, send a message on the session_exit channel.
    A message will be sent on the client_exit channel when ready for client to quit.
    """

    def __init__(
        self,
        cwd: Path,
        paths: List[Path] = [],
        exclude_paths: List[Path] = [],
        ignore_paths: List[Path] = [],
        diff: Optional[str] = None,
        pr_diff: Optional[str] = None,
        config: Config = Config(),
        # Set to false for clients that apply the edits themselves (like vscode)
        apply_edits: bool = True,
    ):
        # All errors thrown here need to be caught here
        self.stopped = Event()

        if not mentat_dir_path.exists():
            os.mkdir(mentat_dir_path)
        setup_logging()
        sentry_init()
        self.id = uuid4()
        self._tasks: Set[asyncio.Task[None]] = set()

        # Since we can't set the session_context until after all of the singletons are created,
        # any singletons used in the constructor of another singleton must be passed in
        llm_api_handler = LlmApiHandler()

        stream = SessionStream()
        self.stream = stream
        self.stream.start()

        cost_tracker = CostTracker()

        code_context = CodeContext(stream, cwd, diff, pr_diff, ignore_paths)

        code_file_manager = CodeFileManager()

        conversation = Conversation()

        vision_manager = VisionManager()

        agent_handler = AgentHandler()

        auto_completer = AutoCompleter()

        sampler = Sampler()

        session_context = SessionContext(
            cwd,
            stream,
            llm_api_handler,
            cost_tracker,
            config,
            code_context,
            code_file_manager,
            conversation,
            vision_manager,
            agent_handler,
            auto_completer,
            sampler,
        )
        self.ctx = session_context
        SESSION_CONTEXT.set(session_context)
        self.error = None

        # Functions that require session_context
        check_version()
        config.send_errors_to_stream()
        for path in paths:
            code_context.include(path, exclude_patterns=exclude_paths)
        if len(code_context.include_files) == 0 and (diff or pr_diff):
            for file in code_context.diff_context.diff_files():
                code_context.include(file)
        if config.sampler:
            sampler.set_active_diff()

        self.apply_edits = apply_edits

    def _create_task(self, coro: Coroutine[None, None, Any]):
        """Utility method for running a Task in the background"""

        def task_cleanup(task: asyncio.Task[None]):
            self._tasks.remove(task)

        task = asyncio.create_task(coro)
        task.add_done_callback(task_cleanup)
        self._tasks.add(task)

        return task

    def send_file_edits(self, file_edits: List[FileEdit]):
        ctx = SESSION_CONTEXT.get()
        ctx.stream.send(
            [
                {
                    "file_path": str(file_edit.file_path),
                    "new_file_path": (None if not file_edit.rename_file_path else str(file_edit.rename_file_path)),
                    "type": (
                        "creation" if file_edit.is_creation else ("deletion" if file_edit.is_deletion else "edit")
                    ),
                    "new_content": "\n".join(
                        file_edit.get_updated_file_lines(
                            ctx.code_file_manager.file_lines.get(file_edit.file_path, []).copy()
                        )
                    ),
                }
                for file_edit in file_edits
            ],
            channel="model_file_edits",
        )

    async def _main(self):
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        code_context = session_context.code_context
        conversation = session_context.conversation
        code_file_manager = session_context.code_file_manager
        agent_handler = session_context.agent_handler

        # check early for ctags so we can fail fast
        if session_context.config.auto_context_tokens > 0:
            ensure_ctags_installed()

        session_context.llm_api_handler.initialize_client()
        check_model()

        need_user_request = True
        while True:
            await code_context.refresh_context_display()
            try:
                if need_user_request:
                    # Normally, the code_file_manager pushes the edits; but when agent mode is on, we want all
                    # edits made between user input to be collected together.
                    if agent_handler.agent_enabled:
                        code_file_manager.history.push_edits()
                        stream.send(
                            "Use /undo to undo all changes from agent mode since last" " input.",
                            style="success",
                        )
                    message = await collect_input_with_commands()
                    if message.data.strip() == "":
                        continue
                    conversation.add_user_message(message.data)

                parsed_llm_response = await conversation.get_model_response()
                file_edits = [file_edit for file_edit in parsed_llm_response.file_edits if file_edit.is_valid()]
                for file_edit in file_edits:
                    file_edit.resolve_conflicts()
                if file_edits:
                    if session_context.config.revisor:
                        await revise_edits(file_edits)

                    if session_context.config.sampler:
                        session_context.sampler.set_active_diff()

                    self.send_file_edits(file_edits)
                    if self.apply_edits:
                        if not agent_handler.agent_enabled:
                            file_edits, need_user_request = await get_user_feedback_on_edits(file_edits)
                        applied_edits = await code_file_manager.write_changes_to_files(file_edits)
                        stream.send(
                            ("Changes applied." if applied_edits else "No changes applied."),
                            style="input",
                        )
                    else:
                        need_user_request = True

                    if agent_handler.agent_enabled:
                        if parsed_llm_response.interrupted:
                            need_user_request = True
                        else:
                            need_user_request = await agent_handler.add_agent_context()
                else:
                    need_user_request = True
                stream.send(bool(file_edits), channel="edits_complete")
            except SessionExit:
                stream.send(None, channel="client_exit")
                break
            except ReturnToUser:
                need_user_request = True
                continue
            except (
                APITimeoutError,
                RateLimitError,
                BadRequestError,
                PermissionDeniedError,
            ) as e:
                stream.send(f"Error accessing OpenAI API: {e.message}", style="error")
                break

    async def listen_for_session_exit(self):
        await self.stream.recv(channel="session_exit")
        self._main_task.cancel()

    async def listen_for_completion_requests(self):
        ctx = SESSION_CONTEXT.get()

        async for message in self.stream.listen(channel="completion_request"):
            completions = ctx.auto_completer.get_completions(
                message.data,
                command_autocomplete=message.extra.get("command_autocomplete", False),
            )
            # Will intermediary client for vscode serialize/deserialize all messages automatically?
            self.stream.send(completions, channel=f"completion_request:{message.id}")

    async def listen_for_include(self):
        ctx = SESSION_CONTEXT.get()

        async for message in self.stream.listen(channel="include"):
            ctx.code_context.include(message.data)
            await ctx.code_context.refresh_context_display()

    async def listen_for_exclude(self):
        ctx = SESSION_CONTEXT.get()

        async for message in self.stream.listen(channel="exclude"):
            ctx.code_context.exclude(message.data)
            await ctx.code_context.refresh_context_display()

    async def listen_for_clear_conversation(self):
        ctx = SESSION_CONTEXT.get()

        async for _ in self.stream.listen(channel="clear_conversation"):
            ctx.conversation.clear_messages()

    ### lifecycle

    def start(self):
        """Asynchronously start the Session.

        A background asyncio Task will be created to run the startup sequence and run
        the main loop which runs until an Exception or session_exit signal is encountered.
        """

        async def run_main():
            ctx = SESSION_CONTEXT.get()
            try:
                with sentry_sdk.start_transaction(op="mentat_started", name="Mentat Started") as transaction:
                    transaction.set_tag("config", attr.asdict(ctx.config))
                    await self._main()
            except (SessionExit, CancelledError):
                pass
            except (MentatError, UserError) as e:
                self.stream.send(str(e), style="error")
            except Exception as e:
                # All unhandled exceptions end up here
                error = f"Unhandled Exception: {traceback.format_exc()}"
                logging.error(error)
                # Helps us handle errors in tests
                if is_test_environment():
                    print(error)
                self.error = error
                sentry_sdk.capture_exception(e)
                self.stream.send(error, style="error")
            finally:
                await self._stop()
                sentry_sdk.flush()

        self._main_task: Task[None] = asyncio.create_task(run_main())

        self._create_task(self.listen_for_session_exit())
        self._create_task(self.listen_for_completion_requests())
        self._create_task(self.listen_for_include())
        self._create_task(self.listen_for_exclude())
        self._create_task(self.listen_for_clear_conversation())

    async def _stop(self):
        if self.stopped.is_set():
            return
        self.stopped.set()

        session_context = SESSION_CONTEXT.get()
        vision_manager = session_context.vision_manager

        vision_manager.close()
        logging.shutdown()

        for task in self._tasks:
            task.cancel()

        self._main_task.cancel()
        try:
            await self._main_task
        except CancelledError:
            pass

        self.stream.send(None, channel="session_stopped")
        await self.stream.join()
        self.stream.stop()



from __future__ import annotations

from contextvars import ContextVar
from pathlib import Path
from typing import TYPE_CHECKING

import attr

if TYPE_CHECKING:
    from mentat.agent_handler import AgentHandler
    from mentat.auto_completer import AutoCompleter
    from mentat.code_context import CodeContext
    from mentat.code_file_manager import CodeFileManager
    from mentat.config import Config
    from mentat.conversation import Conversation
    from mentat.cost_tracker import CostTracker
    from mentat.llm_api_handler import LlmApiHandler
    from mentat.sampler.sampler import Sampler
    from mentat.session_stream import SessionStream
    from mentat.vision.vision_manager import VisionManager

SESSION_CONTEXT: ContextVar[SessionContext] = ContextVar("mentat:session_context")


@attr.define()
class SessionContext:
    cwd: Path = attr.field()
    stream: SessionStream = attr.field()
    llm_api_handler: LlmApiHandler = attr.field()
    cost_tracker: CostTracker = attr.field()
    config: Config = attr.field()
    code_context: CodeContext = attr.field()
    code_file_manager: CodeFileManager = attr.field()
    conversation: Conversation = attr.field()
    vision_manager: VisionManager = attr.field()
    agent_handler: AgentHandler = attr.field()
    auto_completer: AutoCompleter = attr.field()
    sampler: Sampler = attr.field()



import os
from pathlib import Path
from typing import Optional
from uuid import uuid4

from git import GitCommandError, Repo  # type: ignore

from mentat.errors import SampleError
from mentat.git_handler import get_non_gitignored_files
from mentat.utils import is_file_text_encoded

CLONE_TO_DIR = Path("benchmarks/benchmark_repos")


def clone_repo(url: str, local_dir_name: str, refresh: bool = False, depth: int = 0) -> Path | None:
    local_dir = CLONE_TO_DIR / local_dir_name
    if os.path.exists(local_dir):
        if refresh:
            repo = Repo(local_dir)
            repo.git.reset("--hard")
            repo.git.clean("-fd")
            repo.git.fetch("--all")
    else:
        if depth > 0:
            repo = Repo.clone_from(url, local_dir, depth=depth)
        else:
            repo = Repo.clone_from(url, local_dir)
    return local_dir


def apply_diff_to_repo(diff: str, repo: Repo, commit: bool = False) -> str | None:
    """Apply a git diff to a repo. If commit is True, commit the changes."""
    temp_id = uuid4().hex
    try:
        # Save self.diff_merge_base to a temporary .diff file
        with open(f".sample_{temp_id}.diff", "w") as f:
            f.write(diff)
        repo.git.execute(["git", "apply", f".sample_{temp_id}.diff"])
        os.remove(f".sample_{temp_id}.diff")
        if commit:
            repo.git.add(".")
            repo.git.commit("-m", f"sample_{temp_id}")
    except GitCommandError as e:
        try:
            os.remove(f".sample_{temp_id}.diff")
        except FileNotFoundError:
            pass
        return str(e)


def setup_repo(
    url: str,
    cwd: Path | str | None = None,
    depth: int = 0,
    commit: Optional[str] = None,
    diff_merge_base: Optional[str] = None,
    diff_active: Optional[str] = None,
) -> Repo:
    # Locate or clone repo
    repo_name = url.split("/")[-1]
    if cwd is None:
        cwd = clone_repo(
            url=url,
            local_dir_name=repo_name,
            refresh=False,  # Do it below
            depth=depth,
        )
        if cwd is None:
            raise SampleError(f"Error cloning {url}")
    else:
        cwd = Path(cwd)
        if not cwd.exists():
            raise SampleError(f"Error: {cwd} does not exist")
    os.chdir(cwd)

    # Setup git history
    repo = Repo(".")
    repo.git.reset("--hard")
    repo.git.clean("-fd")
    repo.git.fetch("--all")
    if commit is not None:
        repo.git.checkout(commit)
    if diff_merge_base:
        errors = apply_diff_to_repo(diff_merge_base, repo, commit=True)
        if errors:
            raise SampleError(f"Error applying diff_merge_base: {errors}")
    if diff_active:
        errors = apply_diff_to_repo(diff_active, repo)
        if errors:
            raise SampleError(f"Error applying diff_active: {errors}")

    return repo


def get_active_snapshot_commit(repo: Repo) -> str | None:
    """Returns the commit hash of the current active snapshot, or None if there are no active changes."""
    if not repo.is_dirty() and not repo.untracked_files:
        return None
    if not repo.config_reader().has_option("user", "name"):
        raise SampleError("ERROR: Git user.name not set. Please run 'git config --global user.name" ' "Your Name"\'.')
    try:
        # Stash active changes and record the current position
        for file in get_non_gitignored_files(Path(repo.working_dir)):
            if is_file_text_encoded(file):
                repo.git.add(file)
        repo.git.stash("push", "-u")
        detached_head = repo.head.is_detached
        if detached_head:
            current_state = repo.head.commit.hexsha
        else:
            current_state = repo.active_branch.name
        # Commit them on a temporary branch
        temp_branch = f"sample_{uuid4().hex}"
        repo.git.checkout("-b", temp_branch)
        repo.git.stash("apply")
        repo.git.commit("-am", temp_branch)
        # Save the commit hash for diffing against later
        new_commit = repo.head.commit.hexsha
        # Reset repo to how it was before
        repo.git.checkout(current_state)
        repo.git.branch("-D", temp_branch)
        repo.git.stash("apply")
        repo.git.stash("drop")
        # Return the hexsha of the new commit
        return new_commit

    except Exception as e:
        raise SampleError(
            "WARNING: Mentat encountered an error while making temporary git changes:"
            f" {e}. If your active changes have disappeared, they can most likely be "
            "recovered using 'git stash pop'."
        )



from pathlib import Path
from typing import List

from typing_extensions import override

from mentat.auto_completer import get_command_filename_completions
from mentat.command.command import Command, CommandArgument
from mentat.session_context import SESSION_CONTEXT
from mentat.utils import get_relative_path


class ExcludeCommand(Command, command_name="exclude"):
    @override
    async def apply(self, *args: str) -> None:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        code_context = session_context.code_context

        if len(args) == 0:
            stream.send("No files specified", style="warning")
            return
        for file_path in args:
            excluded_paths = code_context.exclude(file_path)
            for excluded_path in excluded_paths:
                rel_path = get_relative_path(excluded_path, session_context.cwd)
                stream.send(f"{rel_path} removed from context", style="error")

    @override
    @classmethod
    def arguments(cls) -> List[CommandArgument]:
        return [CommandArgument("required", ["path", "glob pattern"], repeatable=True)]

    @override
    @classmethod
    def argument_autocompletions(cls, arguments: list[str], argument_position: int) -> list[str]:
        ctx = SESSION_CONTEXT.get()

        file_names = get_command_filename_completions(arguments[-1])
        filtered_file_names: List[str] = []
        for file_name in file_names:
            file_path = Path(file_name).expanduser().resolve()
            for included_file in ctx.code_context.include_files.keys():
                if included_file.is_relative_to(file_path):
                    filtered_file_names.append(file_name)
                    break

        return filtered_file_names

    @override
    @classmethod
    def help_message(cls) -> str:
        return "Remove files from context."



#!/usr/bin/env python
import asyncio
import multiprocessing
import os
from datetime import datetime
from functools import partial
from pathlib import Path

import tqdm
from openai import BadRequestError

from benchmarks.arg_parser import common_benchmark_parser
from benchmarks.benchmark_result import BenchmarkResult
from benchmarks.benchmark_run import BenchmarkRun
from benchmarks.exercise_runners.exercise_runner_factory import ExerciseRunnerFactory
from mentat import Mentat
from mentat.config import Config
from mentat.git_handler import get_mentat_branch, get_mentat_hexsha
from mentat.sampler.utils import clone_repo
from mentat.session_context import SESSION_CONTEXT


def clone_exercism_repo(refresh_repo, language):
    exercism_url = f"https://github.com/exercism/{language}.git"
    local_dir = clone_repo(exercism_url, f"exercism-{language}", refresh_repo)
    os.chdir(local_dir)


prompt = (
    "You are a professional code reviewer who helps other coders improve their skills."
    " You recently assigned a coder a small coding test to assess their level, with a"
    " pre-written stub template and an automated test suite to determine if they"
    " succeeded. They failed the test; using the instructions for the test, the code"
    " they wrote for the test, and the output of the test suite, your job is to"
    " determine why they failed. Give a terse but informative couple of sentences on"
    " why they failed, before giving the reason it failed on the final line in the"
    " format:\n"
    + "reason: <reason_failed>\n"
    + "Your response will be parsed programmatically, so you MUST follow the format for"
    " the final line! The possible responses for the final line and what they mean"
    " are as follows:\n" + "blank (the coder didn't change the file at all from the stub you provided"
    " them)\n" + "wording (everything was correct, but the coder messed up the wording or spacing"
    " which caused it to be rejected)\n"
    + "duplication (the coder had a random duplicated line that caused the code to not"
    " be compiled/interpreted)\n" + "syntax (the coder messed up their syntax, meaning their code couldn't be"
    " compiled/interpreted)\n"
    + "logic (the coder messed up the logic)\n"
    + "other (some other reason caused it to fail)\n"
)
model = "gpt-4-0314"


async def failure_analysis(exercise_runner, language):
    instructions = exercise_runner.read_instructions()
    code = exercise_runner.read_code(language)
    test_results = exercise_runner.read_test_results()

    final_message = f"All instructions:\n{instructions}\nCode to review:\n{code}\nTest" f" results:\n{test_results}"
    messages = [
        {"role": "system", "content": prompt},
        {"role": "user", "content": final_message},
    ]
    response = ""
    try:
        llm_api_handler = SESSION_CONTEXT.get().llm_api_handler
        llm_grade = await llm_api_handler.call_llm_api(messages, model, False)
        response = llm_grade.choices[0].message.content
    except BadRequestError:
        response = "Unable to analyze test case\nreason: too many tokens to analyze"

    try:
        reason = response.strip().split("\n")[-1][8:]
    except IndexError:
        reason = "response error"
    return response, reason


async def run_exercise(problem_dir, language="python", max_iterations=2):
    exercise_runner = ExerciseRunnerFactory.create(language, problem_dir)
    client = Mentat(
        cwd=Path("."),
        paths=exercise_runner.include_files(),
        exclude_paths=exercise_runner.exclude_files(),
        config=Config(),
    )
    await client.startup()

    # These prompts are copied from Aider: https://aider.chat/docs/benchmarks.html to
    # allow for direct comparison.
    prompt_1 = (
        f"Use the instructions in {exercise_runner.docs()} to modify"
        + f" {exercise_runner.file}. Keep and implement the existing"
        + " function or class stubs, they will be called from unit tests. Only use"
        + f" standard {language} libraries, don't suggest installing any packages."
    )
    prompt_2 = (
        "\nSee the testing errors above. The tests are correct. Fix the code"
        + f" in {exercise_runner.file} to resolve the errors."
    )

    iterations = 0
    while iterations < max_iterations and not client.stopped.is_set():
        if exercise_runner.passed():
            break
        message = prompt_1 if iterations == 0 else exercise_runner.get_error_message() + prompt_2
        await client.call_mentat_auto_accept(message)

        exercise_runner.run_test()
        iterations += 1

    had_error = client.stopped.is_set()
    messages = client.get_conversation().literal_messages
    await client.shutdown()
    passed = exercise_runner.passed()
    cost_tracker = SESSION_CONTEXT.get().cost_tracker
    result = BenchmarkResult(
        iterations=iterations,
        passed=passed,
        name=exercise_runner.name,
        tokens=cost_tracker.total_tokens,
        cost=cost_tracker.total_cost,
        transcript={"id": problem_dir, "messages": messages},
    )
    if had_error:
        result.response = "Error while running mentat"
        result.reason = "error"
    elif not result.passed:
        response, reason = await failure_analysis(exercise_runner, language)
        result.response = response
        result.reason = reason
    return result


def run_exercise_sync(problem_dir, language="python", max_iterations=2):
    exercise_runner = ExerciseRunnerFactory.create(language, problem_dir)
    old_result = exercise_runner.get_result_from_txt()
    if old_result:
        return old_result
    try:
        result = asyncio.run(run_exercise(problem_dir, language, max_iterations))
    except Exception as e:
        print(f"\nError running {problem_dir}")
        print(str(e), flush=True)
        result = BenchmarkResult(
            iterations=0,
            passed=False,
            name=problem_dir,
            tokens=0,
            cost=0,
            response=str(e),
            reason="error",
            transcript={"id": problem_dir, "messages": []},
        )
    result.instructions = exercise_runner.read_instructions()
    result.code = exercise_runner.read_code(language)
    result.test_output = exercise_runner.read_test_results()
    with open("results.txt", "a") as f:
        f.write(result.to_json())
        f.write("\n")
    return result


def tqdm_summary(results):
    passed_in_n = {}
    failed = 0
    for result in results:
        if result.passed:
            iteration = result.iterations
            if iteration:
                passed_in_n[iteration] = passed_in_n.get(iteration, 0) + 1
        else:
            failed += 1
    return "Passed: " + str(passed_in_n)[1:-1] + "| Failed: " + str(failed)


def run_exercism_benchmark(
    benchmarks,
    max_benchmarks,
    max_iterations,
    max_workers,
    language,
):
    all_exercises = os.listdir("exercises/practice")
    if len(benchmarks) > 0:
        exercises = list(set(benchmarks) & set(all_exercises))
    else:
        exercises = all_exercises[:max_benchmarks]
    exercises = sorted(exercises)
    num_exercises = len(exercises)

    # TODO: aiomultiprocessing would be faster with fewer workers; setup a Manager in a parent process
    # that controls the children processes so that we don't run into rate limits
    with multiprocessing.Pool(processes=max_workers) as pool:
        pbar = tqdm.tqdm(total=num_exercises)

        result_map = pool.imap(
            partial(run_exercise_sync, language=language, max_iterations=max_iterations),
            exercises,
        )
        results = []
        for result in result_map:
            pbar.update()
            results.append(result)
            pbar.set_description(tqdm_summary(results) + "| Last Ran: " + result.name)
        results.sort(key=lambda result: result.name)

        benchmark_run = BenchmarkRun(
            results,
            metadata={
                "language": language,
                "type": "exercism",
                "date": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                "commit": get_mentat_hexsha(),
                "branch": get_mentat_branch(),
            },
        )
        benchmark_run.save()
        benchmark_run.render_results()


if __name__ == "__main__":
    parser = common_benchmark_parser()
    args = parser.parse_args()
    clone_exercism_repo(args.refresh_repo, args.language)
    print(args)
    run_exercism_benchmark(
        args.benchmarks,
        args.max_benchmarks,
        args.max_iterations,
        args.max_workers,
        args.language,
    )



from __future__ import annotations

import asyncio
import json
import logging
import subprocess
from typing import List, Optional

from openai import RateLimitError
from openai.types.chat import (
    ChatCompletionAssistantMessageParam,
    ChatCompletionContentPartParam,
    ChatCompletionMessageParam,
    ChatCompletionSystemMessageParam,
    ChatCompletionUserMessageParam,
)

from mentat.llm_api_handler import (
    TOKEN_COUNT_WARNING,
    count_tokens,
    get_max_tokens,
    prompt_tokens,
    raise_if_context_exceeds_max,
)
from mentat.parsers.file_edit import FileEdit
from mentat.parsers.parser import ParsedLLMResponse
from mentat.session_context import SESSION_CONTEXT
from mentat.transcripts import ModelMessage, TranscriptMessage, UserMessage
from mentat.utils import add_newline


class MentatAssistantMessageParam(ChatCompletionAssistantMessageParam):
    parsed_llm_response: ParsedLLMResponse


class Conversation:
    def __init__(self):
        self._messages = list[ChatCompletionMessageParam]()

        # This contains a list of messages used for transcripts
        self.literal_messages = list[TranscriptMessage]()

    # The transcript logger logs tuples containing the actual message sent by the user or LLM
    # and (for LLM messages) the LLM conversation that led to that LLM response
    def add_transcript_message(self, transcript_message: TranscriptMessage):
        transcript_logger = logging.getLogger("transcript")
        transcript_logger.info(json.dumps(transcript_message))
        self.literal_messages.append(transcript_message)

    def add_user_message(self, message: str, image: Optional[str] = None):
        """Used for actual user input messages"""
        content: List[ChatCompletionContentPartParam] | str = message
        if image:
            content = [
                {
                    "type": "text",
                    "text": message,
                },
                {
                    "type": "image_url",
                    "image_url": {
                        "url": image,
                    },
                },
            ]
        self.add_transcript_message(UserMessage(message=content, prior_messages=None))
        self.add_message(ChatCompletionUserMessageParam(role="user", content=content))

    def add_model_message(
        self,
        message: str,
        messages_snapshot: list[ChatCompletionMessageParam],
        parsed_llm_response: ParsedLLMResponse,
    ):
        """Used for actual model output messages"""
        self.add_transcript_message(ModelMessage(message=message, prior_messages=messages_snapshot))
        self.add_message(
            MentatAssistantMessageParam(
                parsed_llm_response=parsed_llm_response,
                role="assistant",
                content=message,
            )
        )

    def add_message(self, message: ChatCompletionMessageParam):
        """Used for adding messages to the models conversation. Does not add a left-side message to the transcript!"""
        self._messages.append(message)

    async def count_tokens(
        self,
        system_prompt: Optional[list[ChatCompletionMessageParam]] = None,
        include_code_message: bool = False,
    ) -> int:
        _messages = await self.get_messages(system_prompt=system_prompt, include_code_message=include_code_message)
        model = SESSION_CONTEXT.get().config.model
        return prompt_tokens(_messages, model)

    async def get_messages(
        self,
        system_prompt: Optional[list[ChatCompletionMessageParam]] = None,
        include_parsed_llm_responses: bool = False,
        include_code_message: bool = False,
    ) -> list[ChatCompletionMessageParam]:
        """Returns the messages in the conversation. The system message may change throughout
        the conversation and messages may contain additional metadata not supported by the API,
        so it is important to access the messages through this method.
        """
        ctx = SESSION_CONTEXT.get()

        _messages = [
            (  # Remove metadata from messages by default
                ChatCompletionAssistantMessageParam(role=msg["role"], content=msg.get("content"))
                if msg["role"] == "assistant" and include_parsed_llm_responses is False
                else msg
            )
            for msg in self._messages.copy()
        ]

        if len(_messages) > 0 and _messages[-1].get("role") == "user":
            prompt = _messages[-1].get("content")
            if isinstance(prompt, list):
                text_prompts = [p.get("text", "") for p in prompt if p.get("type") == "text"]
                prompt = " ".join(text_prompts)
        else:
            prompt = ""

        if include_code_message:
            code_message = await ctx.code_context.get_code_message(
                prompt_tokens(_messages, ctx.config.model),
                prompt=(
                    prompt  # Prompt can be image as well as text
                    if isinstance(prompt, str)
                    else ""
                ),
            )
            _messages = [
                ChatCompletionSystemMessageParam(
                    role="system",
                    content=code_message,
                )
            ] + _messages

        if system_prompt is None:
            if ctx.config.no_parser_prompt:
                system_prompt = []
            else:
                parser = ctx.config.parser
                system_prompt = [
                    ChatCompletionSystemMessageParam(
                        role="system",
                        content=parser.get_system_prompt(),
                    )
                ]

        return system_prompt + _messages

    def clear_messages(self) -> None:
        """Clears the messages in the conversation"""
        self._messages = list[ChatCompletionMessageParam]()

    async def _stream_model_response(
        self,
        messages: list[ChatCompletionMessageParam],
    ) -> ParsedLLMResponse:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        code_file_manager = session_context.code_file_manager
        config = session_context.config
        parser = config.parser
        llm_api_handler = session_context.llm_api_handler
        cost_tracker = session_context.cost_tracker

        stream.send(
            None,
            channel="loading",
        )
        response = await llm_api_handler.call_llm_api(
            messages,
            config.model,
            stream=True,
            response_format=parser.response_format(),
        )
        stream.send(
            None,
            channel="loading",
            terminate=True,
        )

        num_prompt_tokens = prompt_tokens(messages, config.model)
        stream.send(f"Total token count: {num_prompt_tokens}", style="info")
        if num_prompt_tokens > TOKEN_COUNT_WARNING:
            stream.send(
                "Warning: LLM performance drops off rapidly at large context sizes. Use"
                " /clear to clear context or use /exclude to exclude any uneccessary"
                " files.",
                style="warning",
            )

        stream.send("Streaming...\n")
        async with stream.interrupt_catcher(parser.shutdown):
            parsed_llm_response = await parser.stream_and_parse_llm_response(add_newline(response))
        # Sampler and History require previous_file_lines
        for file_edit in parsed_llm_response.file_edits:
            file_edit.previous_file_lines = code_file_manager.file_lines.get(file_edit.file_path, []).copy()
        if not parsed_llm_response.interrupted:
            cost_tracker.display_last_api_call()
        else:
            # Generator doesn't log the api call if we interrupt it
            cost_tracker.log_api_call_stats(
                num_prompt_tokens,
                count_tokens(parsed_llm_response.full_response, config.model, full_message=False),
                config.model,
                display=True,
            )

        messages.append(
            ChatCompletionAssistantMessageParam(role="assistant", content=parsed_llm_response.full_response)
        )
        self.add_model_message(parsed_llm_response.full_response, messages, parsed_llm_response)

        return parsed_llm_response

    async def get_model_response(self) -> ParsedLLMResponse:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        config = session_context.config

        messages_snapshot = await self.get_messages(include_code_message=True)
        tokens_used = prompt_tokens(messages_snapshot, config.model)
        raise_if_context_exceeds_max(tokens_used)

        try:
            response = await self._stream_model_response(messages_snapshot)
        except RateLimitError:
            stream.send(
                "Rate limit error received from OpenAI's servers using model"
                f' {config.model}.\nUse "/config model <model_name>" to switch to a'
                " different model.",
                style="error",
            )
            return ParsedLLMResponse("", "", list[FileEdit]())
        return response

    async def remaining_context(self) -> int | None:
        ctx = SESSION_CONTEXT.get()
        return get_max_tokens() - prompt_tokens(await self.get_messages(), ctx.config.model)

    async def can_add_to_context(self, message: str) -> bool:
        """
        Whether or not the model has enough context remaining to add this message.
        Will take token buffer into account and uses full_message=True.
        """
        ctx = SESSION_CONTEXT.get()

        remaining_context = await self.remaining_context()
        return (
            remaining_context is not None
            and remaining_context - count_tokens(message, ctx.config.model, full_message=True) - ctx.config.token_buffer
            > 0
        )

    async def run_command(self, command: list[str]) -> bool:
        """
        Runs a command and, if there is room, adds the output to the conversation under the 'system' role.
        """
        ctx = SESSION_CONTEXT.get()

        ctx.stream.send("Running command: ", end="", style="info")
        ctx.stream.send(" ".join(command), style="warning")
        ctx.stream.send("Command output:", style="info")

        try:
            process = subprocess.Popen(
                command,
                cwd=ctx.cwd,
                text=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
            output = list[str]()
            while True:
                if process.stdout is None:
                    break
                line = process.stdout.readline()
                if not line:
                    break
                output.append(line)
                ctx.stream.send(line, end="")
                # This gives control back to the asyncio event loop so we can actually print what we sent
                # Unfortunately asyncio.sleep(0) won't work https://stackoverflow.com/a/74505785
                # Note: if subprocess doesn't flush, output can't and won't be streamed.
                await asyncio.sleep(0.01)
        except FileNotFoundError:
            output = [f"Invalid command: {' '.join(command)}"]
            ctx.stream.send(output[0])
        output = "".join(output)
        message = f"Command ran:\n{' '.join(command)}\nCommand output:\n{output}"

        if await self.can_add_to_context(message):
            self.add_message(ChatCompletionSystemMessageParam(role="system", content=message))
            ctx.stream.send("Successfully added command output to model context.", style="success")
            return True
        else:
            ctx.stream.send(
                "Not enough tokens remaining in model's context to add command output" " to model context.",
                style="error",
            )
            return False

    def _get_user_message(self, message: ChatCompletionUserMessageParam) -> str:
        if not message["content"]:
            return ""
        elif isinstance(message["content"], str):
            return message["content"]
        else:
            full = ""
            for part in message["content"]:
                if part["type"] == "text":
                    full += part["text"]
            return full

    def amend(self) -> Optional[str]:
        for i, message in reversed(list(enumerate(self._messages))):
            if message["role"] == "user" and self._get_user_message(message):
                self._messages = self._messages[:i]
                return self._get_user_message(message)



import os
import subprocess
from pathlib import Path

from benchmarks.benchmark_result import BenchmarkResult


class AbstractExerciseRunner:
    def __init__(self, exercise, extension):
        self.name = exercise
        self.dir = Path(f"exercises/practice/{exercise}").absolute()
        self.file = Path(f"{exercise}.{extension}")
        self.full_path = self.dir / self.file
        self.test_output_file = self.dir / "test_output.txt"

    def _run_test_command(self, command, cwd="."):
        try:
            proc = subprocess.run(
                command,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                timeout=5,
                cwd=cwd,
            )
            results = proc.stdout.decode("utf-8")
        except subprocess.TimeoutExpired:
            results = "Test timed out"
        with open(self.test_output_file, "w") as f:
            f.write(results.strip())

    def include_files(self):
        return [self.full_path, self.docs()]

    def exclude_files(self):
        return [self.docs() / "hints.md"]

    def docs(self):
        return self.dir / ".docs"

    def get_result_from_txt(self):
        if not Path("results.txt").exists():
            return None
        with open("results.txt", "r") as f:
            for line in f.readlines():
                if f'"{self.name}"' in line:
                    return BenchmarkResult.load_json(line)

    def get_error_message(self):
        with open(self.test_output_file, "r") as f:
            lines = f.readlines()
            lines = lines[:50]
            return "\n".join(lines)

    # This will include hint.md
    def read_instructions(self):
        instructions = ""
        for file_name in os.listdir(self.docs()):
            with open(self.docs() / file_name) as f:
                contents = f.read()
                instructions += f"{file_name}\n{contents}\n"
        return instructions

    def read_code(self, language):
        code = ""
        with open(self.full_path) as f:
            contents = f.read()
            code += f"{self.file}\n{contents}"
        return code

    def read_test_results(self):
        if not self.test_output_file.exists():
            return ""
        with open(self.test_output_file) as f:
            contents = f.read()
            test_results = f"test_output.txt\n{contents}"
        return test_results



class ModelError(Exception):
    """
    Raised when model output doesn't adhere to the specified format for changes.
    Handled by the parser; shouldn't be thrown outside of parsing!
    """


# TODO: Combine MentatError and UserError into just MentatError
class MentatError(Exception):
    """
    Will show the user the exception, but not the stacktrace. Used for known errors.
    """


class UserError(Exception):
    """
    Will show the user the exception, but not the stacktrace. Used for known errors.
    """


class HistoryError(Exception):
    """
    Raised when an execption is encountered undoing or redoing edits.
    """


class SampleError(Exception):
    """
    Raised when an exception is raised by a Sample.
    """


class SessionExit(Exception):
    """
    Stops the session without any sign of error.
    """


class PathValidationError(Exception):
    pass


class ReturnToUser(Exception):
    """
    Raised when trying to call the API with too many tokens for that model.
    Will give control back to the user after being raised.
    """



from enum import Enum
from pathlib import Path

from typing_extensions import override

from mentat.code_file_manager import CodeFileManager
from mentat.parsers.change_display_helper import (
    DisplayInformation,
    get_file_action_type,
    highlight_text,
)
from mentat.parsers.diff_utils import matching_index
from mentat.parsers.file_edit import FileEdit, Replacement
from mentat.parsers.parser import Parser
from mentat.parsers.streaming_printer import FormattedString
from mentat.prompts.prompts import read_prompt

unified_diff_parser_prompt_filename = Path("unified_diff_parser_prompt.txt")


class UnifiedDiffDelimiter(Enum):
    SpecialStart = "---"
    SpecialEnd = "+++"
    MidChange = "@@ @@\n"
    EndChange = "@@ end @@\n"


class UnifiedDiffParser(Parser):
    @override
    def get_system_prompt(self) -> str:
        return read_prompt(unified_diff_parser_prompt_filename)

    @override
    def provide_line_numbers(self) -> bool:
        return False

    @override
    def _code_line_beginning(self, display_information: DisplayInformation, cur_block: str) -> FormattedString:
        return ("", {})

    @override
    def _code_line_content(
        self,
        display_information: DisplayInformation,
        content: str,
        cur_line: str,
        cur_block: str,
    ) -> FormattedString:
        if cur_line == UnifiedDiffDelimiter.MidChange.value:
            return [("", {"delimiter": True}), ("\n", {})]
        elif cur_line.startswith("+"):
            return (content, {"color": "green"})
        elif cur_line.startswith("-"):
            return (content, {"color": "red"})
        else:
            return highlight_text(content, display_information.lexer)

    @override
    def _could_be_special(self, cur_line: str) -> bool:
        return (
            # Since the model is printing the context lines, we can only
            # highlight them once we get a full line, so we choose to
            # add the lines to the printer all at once.
            not cur_line.endswith("\n")
            or UnifiedDiffDelimiter.EndChange.value.startswith(cur_line)
            or cur_line.startswith(UnifiedDiffDelimiter.SpecialStart.value)
            or cur_line.startswith(UnifiedDiffDelimiter.SpecialEnd.value)
            or UnifiedDiffDelimiter.SpecialStart.value.startswith(cur_line)
            or UnifiedDiffDelimiter.SpecialEnd.value.startswith(cur_line)
        )

    @override
    def _starts_special(self, line: str) -> bool:
        return line.startswith(UnifiedDiffDelimiter.SpecialStart.value)

    @override
    def _ends_special(self, line: str) -> bool:
        return line.startswith(UnifiedDiffDelimiter.MidChange.value.strip()) or line.startswith(
            UnifiedDiffDelimiter.EndChange.value.strip()
        )

    @override
    def _special_block(
        self,
        code_file_manager: CodeFileManager,
        cwd: Path,
        rename_map: dict[Path, Path],
        special_block: str,
    ) -> tuple[DisplayInformation, FileEdit, bool]:
        lines = special_block.strip().split("\n")
        file_name = lines[0][4:]
        new_name = lines[1][4:]
        is_creation = file_name == "/dev/null"
        is_deletion = new_name == "/dev/null"
        if is_creation:
            file_name = new_name
        if file_name == new_name or is_deletion:
            new_name = None
        else:
            new_name = Path(new_name)
        file_name = Path(file_name)
        full_path = (cwd / file_name).resolve()
        file_lines = self._get_file_lines(code_file_manager, rename_map, full_path)
        file_action_type = get_file_action_type(is_creation, is_deletion, new_name)
        display_information = DisplayInformation(file_name, file_lines, [], [], file_action_type, -1, -1, new_name)
        file_edit = FileEdit(
            full_path,
            [],
            is_creation,
            is_deletion,
            (cwd / new_name).resolve() if new_name else None,
        )
        return (
            display_information,
            file_edit,
            lines[-1].startswith(UnifiedDiffDelimiter.MidChange.value.strip()),
        )

    @override
    def _ends_code(self, line: str) -> bool:
        return line.strip() == UnifiedDiffDelimiter.EndChange.value.strip()

    @override
    def _add_code_block(
        self,
        code_file_manager: CodeFileManager,
        rename_map: dict[Path, Path],
        special_block: str,
        code_block: str,
        display_information: DisplayInformation,
        file_edit: FileEdit,
    ):
        file_lines = self._get_file_lines(code_file_manager, rename_map, file_edit.file_path).copy()

        # First, we split by the symbols that separate changes.
        lines = code_block.split("\n")
        changes = list[list[str]]()
        cur_lines = list[str]()
        for line in lines:
            if (
                line.strip() == UnifiedDiffDelimiter.MidChange.value.strip()
                or line.strip() == UnifiedDiffDelimiter.EndChange.value.strip()
            ):
                changes.append(cur_lines)
                cur_lines = list[str]()
                if line.strip() == UnifiedDiffDelimiter.EndChange.value.strip():
                    break
            else:
                if (
                    # Remove empty lines; hopefully the model always puts a space for context lines
                    line and not line.startswith("+") and not line.startswith("-") and not line.startswith(" ")
                ):
                    return
                cur_lines.append(line)
        if cur_lines:
            changes.append(cur_lines)

        # Next, we collect the - and context lines, search for their locations, and set the replacement ranges
        replacements = list[Replacement]()
        for change in changes:
            if not change:
                continue
            # We need both removals and context in this array this one
            search_lines = list[str]()
            for line in change:
                if line.startswith("-") or line.startswith(" "):
                    search_lines.append(line[1:])
            if not search_lines:
                # If the model gave us no context lines, we place at the start of the file;
                # this most commonly happens with imports
                replacements.append(Replacement(0, 0, [line[1:] for line in change]))
                continue

            start_index = matching_index(file_lines, search_lines)
            if start_index == -1:
                return

            # Matching lines checks for matches that are missing whitespace only lines;
            # this will cause errors with line numbering if we don't add those lines into the change lines
            cur_file_index = start_index
            cur_change_index = 0
            while cur_change_index < len(change) and cur_file_index < len(file_lines):
                cur_change_line = change[cur_change_index]
                if cur_change_line.startswith("+"):
                    cur_change_index += 1
                    continue
                cur_file_line = file_lines[cur_file_index]

                if not cur_file_line.strip() and cur_change_line.strip():
                    change.insert(cur_change_index, "")
                elif not cur_change_line.strip() and cur_file_line.strip():
                    file_lines.insert(cur_file_index, "")
                cur_file_index += 1
                cur_change_index += 1

            # We need a separate Replacement whenever context lines are between a group of additions/removals
            cur_start = None
            cur_additions = list[str]()
            cur_index = start_index
            for line in change:
                if line.startswith(" ") or not line:
                    if cur_start is not None:
                        replacements.append(Replacement(cur_start, cur_index, cur_additions))
                    cur_index += 1
                    cur_additions = list[str]()
                    cur_start = None
                elif line.startswith("+"):
                    if cur_start is None:
                        cur_start = cur_index
                    cur_additions.append(line[1:])
                elif line.startswith("-"):
                    if cur_start is None:
                        cur_start = cur_index
                    cur_index += 1
            if cur_start is not None:
                replacements.append(Replacement(cur_start, cur_index, cur_additions))

        file_edit.replacements.extend(replacements)



from __future__ import annotations

import json
from argparse import ArgumentParser, Namespace
from json import JSONDecodeError
from pathlib import Path

import attr
from attr import converters, validators

from mentat.git_handler import get_git_root_for_path
from mentat.llm_api_handler import known_models
from mentat.parsers.parser import Parser
from mentat.parsers.parser_map import parser_map
from mentat.session_context import SESSION_CONTEXT
from mentat.terminal.themes import themes
from mentat.utils import mentat_dir_path

config_file_name = Path(".mentat_config.json")
user_config_path = mentat_dir_path / config_file_name


def int_or_none(s: str | None) -> int | None:
    if s is not None:
        return int(s)
    return None


bool_autocomplete = ["True", "False"]


@attr.define
class Config:
    _errors: list[str] = attr.field(factory=list)

    # Model specific settings
    model: str = attr.field(
        default="gpt-4-0125-preview",
        metadata={"auto_completions": list(known_models.keys())},
    )
    feature_selection_model: str = attr.field(
        default="gpt-4-1106-preview",
        metadata={"auto_completions": list(known_models.keys())},
    )
    embedding_model: str = attr.field(
        default="text-embedding-ada-002",
        metadata={"auto_completions": [model.name for model in known_models.values() if model.embedding_model]},
    )
    temperature: float = attr.field(default=0.2, converter=float, validator=[validators.le(1), validators.ge(0)])

    maximum_context: int | None = attr.field(
        default=None,
        metadata={
            "description": (
                "The maximum number of lines of context to include in the prompt. It is"
                " inferred automatically for openai models but you can still set it to"
                " save costs. If not set for non-openai models, it defaults to 4096."
            ),
        },
        converter=int_or_none,
        validator=validators.optional(validators.ge(0)),
    )
    token_buffer: int = attr.field(
        default=1000,
        metadata={
            "description": ("The amount of tokens to always be reserved as a buffer for user and" " model messages."),
        },
    )
    parser: Parser = attr.field(  # pyright: ignore
        default="block",
        metadata={
            "description": (
                "The format for the LLM to write code in. You probably don't want to" " mess with this setting."
            ),
            "auto_completions": list(parser_map.keys()),
        },
        converter=parser_map.get,  # pyright: ignore
        validator=validators.instance_of(Parser),  # pyright: ignore
    )
    no_parser_prompt: bool = attr.field(
        default=False,
        metadata={
            "description": (
                "Whether to include the parser prompt in the system message. This"
                " should only be set to true for fine tuned models."
            ),
            "auto_completions": bool_autocomplete,
        },
        converter=converters.optional(converters.to_bool),
    )
    revisor: bool = attr.field(
        default=False,
        metadata={
            "description": (
                "Enables or disables a revisor tweaking model edits after they're made."
                " The revisor will use the same model regular edits do."
            ),
            "auto_completions": bool_autocomplete,
        },
        converter=converters.optional(converters.to_bool),
    )
    sampler: bool = attr.field(
        default=False,
        metadata={
            "description": ("Automatically saves a git diff snapshot for the sampler on startup."),
            "auto_completions": bool_autocomplete,
        },
        converter=converters.optional(converters.to_bool),
    )

    # Context specific settings
    file_exclude_glob_list: list[str] = attr.field(
        factory=list,
        metadata={"description": "List of glob patterns to exclude from context"},
    )
    auto_context_tokens: int = attr.field(  # pyright: ignore
        default=0,
        metadata={
            "description": (
                "Automatically selects code files for every request to include in"
                " context. Adds this many tokens to context each request."
            ),
            "abbreviation": "a",
            "const": 5000,
        },
        converter=int,
        validator=validators.ge(0),  # pyright: ignore
    )
    llm_feature_filter: int = attr.field(  # pyright: ignore
        default=0,
        metadata={
            "description": (
                "Send this many tokens of auto-context-selected code files to an LLM"
                " along with the user_prompt to post-select only files which are"
                " relevant to the task. Post-files will then be sent to the LLM again"
                " to respond to the user's prompt."
            ),
            "abbreviation": "l",
            "const": 5000,
        },
        converter=int,
        validator=validators.ge(0),  # pyright: ignore
    )

    # Sample specific settings
    sample_repo: str | None = attr.field(
        default=None,
        metadata={"description": "A public url for a cloneable git repository to sample from."},
    )
    sample_merge_base_target: str | None = attr.field(
        default=None,
        metadata={"description": "The branch or commit to use as the merge base for samples."},
    )

    theme: str = attr.field(  # pyright: ignore
        default="dark",
        metadata={
            "description": "Theme for interaction possible choices are light or dark.",
            "no_midsession_change": True,
        },
        validator=validators.in_(themes.keys()),  # pyright: ignore
    )

    @classmethod
    def get_fields(cls) -> list[str]:
        return [field.name for field in attr.fields(cls) if not field.name.startswith("_")]

    @classmethod
    def add_fields_to_argparse(cls, parser: ArgumentParser) -> None:
        for field in attr.fields(cls):
            if "no_flag" in field.metadata:
                continue
            name = [f"--{field.name.replace('_', '-')}"]
            if "abbreviation" in field.metadata:
                name.append(f"-{field.metadata['abbreviation'].replace('_', '-')}")

            arguments = {
                "help": field.metadata.get("description", ""),
            }
            if "const" in field.metadata:
                arguments["nargs"] = "?"
                arguments["const"] = field.metadata["const"]

            if field.type == "bool":
                if arguments.get("default", False):
                    arguments["action"] = "store_false"
                else:
                    arguments["action"] = "store_true"
            elif field.type == "int":
                arguments["type"] = int
            elif field.type == "float":
                arguments["type"] = float
            elif field.type == "list[str]":
                arguments["nargs"] = "*"

            parser.add_argument(*name, **arguments)

    @classmethod
    def create(cls, cwd: Path, args: Namespace | None = None) -> Config:
        config = Config()

        # Each method overwrites the previous so they are in order of precedence
        config.load_file(user_config_path)
        git_root = get_git_root_for_path(cwd, raise_error=False)
        if git_root is not None:
            config.load_file(git_root / config_file_name)
        config.load_file(cwd / config_file_name)

        if args is not None:
            config.load_namespace(args)

        return config

    def load_namespace(self, args: Namespace) -> None:
        for field in attr.fields(Config):
            if field.name in args and field.name != "_errors":
                value = getattr(args, field.name)
                if value is not None:
                    try:
                        setattr(self, field.name, value)
                    except (ValueError, TypeError) as e:
                        self.error(f"Warning: Illegal value for {field}: {e}")

    def load_file(self, path: Path) -> None:
        if path.exists():
            with open(path) as config_file:
                try:
                    config = json.load(config_file)
                except JSONDecodeError:
                    self.error(f"Warning: Config {path} contains invalid json; ignoring user" " configuration file")
                    return
            for field in config:
                if hasattr(self, field):
                    try:
                        setattr(self, field, config[field])
                    except (ValueError, TypeError) as e:
                        self.error(f"Warning: Config {path} contains invalid value for" f" setting: {field}\n{e}")
                else:
                    self.error(f"Warning: Config {path} contains unrecognized setting: {field}")

    def error(self, message: str) -> None:
        self._errors.append(message)
        try:
            self.send_errors_to_stream()
        except LookupError:
            pass

    def send_errors_to_stream(self):
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        for error in self._errors:
            stream.send(error, style="warning")
        self._errors = []



import subprocess
from pathlib import Path
from uuid import uuid4

from git import GitCommandError, Repo  # type: ignore

from mentat.code_feature import get_consolidated_feature_refs
from mentat.errors import SampleError
from mentat.git_handler import get_git_diff, get_git_root_for_path, get_hexsha_active
from mentat.parsers.git_parser import GitParser
from mentat.parsers.parser import ParsedLLMResponse
from mentat.sampler.sample import Sample
from mentat.sampler.utils import get_active_snapshot_commit
from mentat.session_context import SESSION_CONTEXT
from mentat.session_input import collect_user_input
from mentat.utils import get_relative_path


class Sampler:
    diff_active: str | None = None
    commit_active: str | None = None
    last_sample_id: str | None = None
    last_sample_hexsha: str | None = None

    def set_active_diff(self):
        # Create a temporary commit with the active changes
        ctx = SESSION_CONTEXT.get()
        git_root = get_git_root_for_path(ctx.cwd, raise_error=False)
        if not git_root:
            return
        repo = Repo(git_root)
        try:
            self.commit_active = get_active_snapshot_commit(repo)
            # If changes were made since the last sample, don't list it as parent.
            if not self.last_sample_hexsha:
                return
            if self.last_sample_hexsha != get_hexsha_active():
                self.last_sample_id = None
                self.last_sample_hexsha = None
        except SampleError as e:
            ctx.stream.send(
                f"Sampler error setting active diff: {e}. Disabling sampler.",
                style="error",
            )
            ctx.config.sampler = False

    async def create_sample(self) -> Sample:
        # Check for repo and merge_base in config
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        code_context = session_context.code_context
        config = session_context.config
        conversation = session_context.conversation

        git_root = get_git_root_for_path(session_context.cwd, raise_error=False)
        if not git_root:
            raise SampleError("No git repo found")

        stream.send("Input sample data", style="input")
        git_repo = Repo(git_root)
        merge_base = None
        if config.sample_merge_base_target:
            target = config.sample_merge_base_target
            stream.send(f"Use merge base target from config ({target})? (y/N)")
            response = (await collect_user_input()).data.strip()
            if response == "y":
                try:
                    mb = git_repo.merge_base(git_repo.head.commit, target)[0]
                    assert mb and hasattr(mb, "hexsha"), "No merge base found"
                    merge_base = mb.hexsha
                except Exception as e:
                    stream.send(f"Error getting merge base from tar: {e}", style="error")
        if not merge_base:
            merge_base = git_repo.head.commit.hexsha
            stream.send(
                f"Use latest commit ({merge_base[:10]} as merge base? Press 'ENTER' to"
                " accept, or enter a new merge base commit."
            )
            response = str((await collect_user_input()).data).strip()
            if response:
                merge_base = response
        try:
            assert merge_base is not None, "No merge base found"
            diff_merge_base = get_git_diff(merge_base, "HEAD")
        except (AssertionError, GitCommandError) as e:
            raise SampleError(f"Error getting diff for merge base: {e}")

        repo = config.sample_repo
        if not repo:
            remote_url = ""
            try:
                remote_url = subprocess.check_output(
                    ["git", "config", "--get", "remote.origin.url"],
                    text=True,
                ).strip()
            except subprocess.CalledProcessError:
                pass
            stream.send(f"Found repo URL: {remote_url}. Press 'ENTER' to accept, or enter a new" " URL.")
            response = (await collect_user_input()).data.strip()
            if response == "y":
                repo = remote_url
            else:
                repo = str(response)
            config.sample_repo = repo

        stream.send("Sample Title:")
        title = (await collect_user_input()).data.strip() or ""
        stream.send("Description: (optional)")
        description = (await collect_user_input()).data.strip() or ""
        stream.send("Test Command: (optional, e.g. 'pytest -k foo')")
        test_command = (await collect_user_input()).data.strip() or ""

        message_history: list[dict[str, str]] = []
        message_prompt = ""
        response_edit: None | ParsedLLMResponse = None
        messages = await conversation.get_messages(include_parsed_llm_responses=True)
        for m in messages[::-1]:
            response: str | ParsedLLMResponse | None = None
            role, content = m["role"], m.get("content")
            if role == "user":
                if isinstance(content, str):
                    text = content
                elif isinstance(content, list) and len(content) > 0:
                    text = content[0].get("text", "")
                else:
                    continue

                if not message_prompt:
                    message_prompt = text
                else:
                    message_history.insert(0, {"role": role, "content": text})

            elif role == "assistant":
                parsed_llm_response = m.get("parsed_llm_response")
                if parsed_llm_response is None:
                    raise SampleError(
                        "Assistant messages must include a the parsed_llm_response."
                        " Hint: Use"
                        " mentat.conversation.MentatAssistantMessageParam"
                        " instead  of"
                        " openai.types.chat.ChatCompletionAssistantMessageParam."
                    )

                if not message_prompt:
                    response_edit = parsed_llm_response
                else:
                    message_history.append(
                        {
                            "role": role,
                            "content": GitParser().file_edits_to_llm_message(parsed_llm_response),
                        }
                    )

        if not response_edit:
            raise SampleError("No LLM response found.")
        message_edit = response_edit.conversation.strip()
        if self.commit_active:
            diff_active = get_git_diff("HEAD", self.commit_active)
        else:
            diff_active = ""
        if response_edit.file_edits:
            diff_edit = get_git_diff(self.commit_active or "HEAD")
        else:
            diff_edit = ""

        context = set[str]()

        def _rp(f: str | Path) -> str:
            return get_relative_path(Path(f), git_root).as_posix()

        # Add include_files from context
        if code_context.include_files:
            feature_refs = get_consolidated_feature_refs([f for fs in code_context.include_files.values() for f in fs])
            context.update(_rp(f) for f in feature_refs)
        # Undo adds/removes/renames to match pre-diff_edit state
        if diff_edit:
            file_edits = GitParser().parse_llm_response(diff_edit).file_edits
            for file_edit in file_edits:
                file_path = _rp(file_edit.file_path)
                rename_path = "" if not file_edit.rename_file_path else _rp(file_edit.rename_file_path)
                if file_edit.is_deletion or rename_path or file_edit.replacements:
                    context.add(file_path)
                if file_edit.is_creation and file_path in context:
                    context.remove(file_path)
                if rename_path and rename_path in context:
                    context.remove(rename_path)
            # TODO: Prompt User to modify/approve context

        sample = Sample(
            title=title,
            description=description,
            id=uuid4().hex,
            parent_id=self.last_sample_id or "",
            repo=repo,
            merge_base=merge_base,
            diff_merge_base=diff_merge_base,
            diff_active=diff_active,
            message_history=message_history,
            message_prompt=message_prompt,
            message_edit=message_edit,
            context=list(context),
            diff_edit=diff_edit,
            test_command=test_command,
        )

        # Save the hexsha and id
        self.last_sample_id = sample.id
        self.last_sample_hexsha = get_hexsha_active()
        self.commit_active = None

        return sample



import shlex
from pathlib import Path
from typing import List

from openai import BadRequestError
from openai.types.chat import (
    ChatCompletionAssistantMessageParam,
    ChatCompletionMessageParam,
    ChatCompletionSystemMessageParam,
)

from mentat.prompts.prompts import read_prompt
from mentat.session_context import SESSION_CONTEXT
from mentat.session_input import ask_yes_no, collect_user_input
from mentat.transcripts import ModelMessage

agent_file_selection_prompt_path = Path("agent_file_selection_prompt.txt")
agent_command_prompt_path = Path("agent_command_selection_prompt.txt")


class AgentHandler:
    def __init__(self):
        self._agent_enabled = False

        self.agent_file_message = ""
        self.agent_file_selection_prompt = read_prompt(agent_file_selection_prompt_path)
        self.agent_command_prompt = read_prompt(agent_command_prompt_path)

    # Make this property readonly because we have to set things when we enable agent mode
    @property
    def agent_enabled(self):
        return self._agent_enabled

    def disable_agent_mode(self):
        self._agent_enabled = False

    async def enable_agent_mode(self):
        ctx = SESSION_CONTEXT.get()

        ctx.stream.send("Finding files to determine how to test changes...", style="info")
        features = ctx.code_context.get_all_features(split_intervals=False)
        messages: list[ChatCompletionMessageParam] = [
            ChatCompletionSystemMessageParam(role="system", content=self.agent_file_selection_prompt),
            ChatCompletionSystemMessageParam(
                role="system",
                content="\n".join(str(feature.path.relative_to(ctx.cwd)) for feature in features),
            ),
        ]
        model = ctx.config.model
        response = await ctx.llm_api_handler.call_llm_api(messages, model, False)
        content = response.choices[0].message.content or ""

        paths = [Path(path) for path in content.strip().split("\n") if Path(path).exists()]
        self.agent_file_message = ""
        for path in paths:
            file_contents = "\n\n".join(ctx.code_file_manager.read_file(path))
            self.agent_file_message += f"{path}\n\n{file_contents}"

        ctx.stream.send(
            "The model has chosen these files to help it determine how to test its" " changes:",
            style="info",
        )
        ctx.stream.send("\n".join(str(path) for path in paths))
        ctx.cost_tracker.display_last_api_call()

        messages.append(ChatCompletionAssistantMessageParam(role="assistant", content=content))
        ctx.conversation.add_transcript_message(
            ModelMessage(message=content, prior_messages=messages, message_type="agent")
        )
        self._agent_enabled = True

    async def _determine_commands(self) -> List[str]:
        ctx = SESSION_CONTEXT.get()

        model = ctx.config.model
        system_prompt: list[ChatCompletionMessageParam] = [
            ChatCompletionSystemMessageParam(role="system", content=self.agent_command_prompt),
            ChatCompletionSystemMessageParam(role="system", content=self.agent_file_message),
        ]
        messages = await ctx.conversation.get_messages(system_prompt=system_prompt)

        try:
            # TODO: Should this even be a separate call or should we collect commands in the edit call?
            response = await ctx.llm_api_handler.call_llm_api(messages, model, False)
            ctx.cost_tracker.display_last_api_call()
        except BadRequestError as e:
            ctx.stream.send(f"Error accessing OpenAI API: {e.message}", style="error")
            return []

        content = response.choices[0].message.content or ""

        messages.append(ChatCompletionAssistantMessageParam(role="assistant", content=content))
        parsed_llm_response = await ctx.config.parser.parse_llm_response(content)
        ctx.conversation.add_model_message(content, messages, parsed_llm_response)

        commands = content.strip().split("\n")
        return commands

    async def add_agent_context(self) -> bool:
        """
        Returns whether or not control should be handed back to user
        """
        ctx = SESSION_CONTEXT.get()

        commands = await self._determine_commands()
        if not commands:
            return True
        ctx.stream.send("The model has chosen these commands to test its changes:", style="info")
        for command in commands:
            ctx.stream.send("* ", end="")
            ctx.stream.send(command, style="warning")
        ctx.stream.send("Run these commands?", style="info")
        run_commands = await ask_yes_no(default_yes=True)
        if not run_commands:
            ctx.stream.send(
                "Enter a new-line separated list of commands to run, or nothing to" " return control to the user:",
                style="info",
            )
            commands: list[str] = (await collect_user_input()).data.strip().splitlines()
            if not commands:
                return True

        ctx.conversation.add_message(
            ChatCompletionSystemMessageParam(
                role="system",
                content=(
                    "You are currently being run autonomously. The following commands"
                    " are being run to test your previous changes. If the commands show"
                    " any errors with your changes, fix them. In order to return"
                    " control to the user, make no more changes to the files. If you"
                    " don't know how to fix a problem, do not waste time trying to"
                    " solve it! The user would much prefer to regain control if you"
                    " can't solve a problem."
                ),
            )
        )
        for command in commands:
            await ctx.conversation.run_command(shlex.split(command))
        return False



from typing import List

from mentat.errors import HistoryError
from mentat.parsers.file_edit import FileEdit
from mentat.session_context import SESSION_CONTEXT


# TODO: Keep track of when we create directories so we can undo those as well
class EditHistory:
    def __init__(self):
        self.edits = list[list[FileEdit]]()
        self.cur_edits = list[FileEdit]()
        self.undone_edits = list[list[FileEdit]]()

    def add_edit(self, file_edit: FileEdit):
        self.cur_edits.append(file_edit)

    def push_edits(self):
        if self.cur_edits:
            self.edits.append(self.cur_edits)
            self.cur_edits = list[FileEdit]()

    def undo(self) -> List[str]:
        if not self.edits:
            return ["No edits available to undo"]

        # Make sure to go top down
        cur_edit = self.edits.pop()
        errors = list[str]()
        undone_edit = list[FileEdit]()
        while cur_edit:
            cur_file_edit = cur_edit.pop()
            try:
                cur_file_edit.undo()
                undone_edit.append(cur_file_edit)
            except HistoryError as e:
                errors.append(str(e))
        if undone_edit:
            self.undone_edits.append(undone_edit)
        return errors

    async def redo(self) -> List[str]:
        if not self.undone_edits:
            return ["No edits available to redo"]

        session_context = SESSION_CONTEXT.get()
        code_file_manager = session_context.code_file_manager

        edits_to_redo = self.undone_edits.pop()
        edits_to_redo.reverse()
        for edit in edits_to_redo:
            edit.display_full_edit(code_file_manager.file_lines[edit.file_path])
        await code_file_manager.write_changes_to_files(edits_to_redo)
        return []

    def undo_all(self) -> List[str]:
        if not self.edits:
            return ["No edits available to undo"]

        errors = list[str]()
        while self.edits:
            error = self.undo()
            if error:
                errors += error
        return errors



#!/usr/bin/env python
import webbrowser
from pathlib import Path

from jinja2 import Environment, FileSystemLoader, select_autoescape

from benchmarks.benchmark_run_summary import BenchmarkRunSummary
from benchmarks.plot_generator import generate_plot_html


def generate_list(path: Path, output: Path):
    summaries = []
    for file in path.iterdir():
        if file.is_file() and file.suffix == ".json":
            summary = BenchmarkRunSummary.load_file(file)
            summaries.append(summary)

    plot = generate_plot_html(summaries)
    env = Environment(
        loader=FileSystemLoader(
            [
                Path(__file__).parent / "resources/templates",
                Path(__file__).parent.parent / "mentat/resources/templates",
            ]
        ),
        autoescape=select_autoescape(["html", "xml"]),
    )
    template = env.get_template("benchmark_list.jinja")
    rendered_html = template.render(summary_list=summaries, plot=plot)

    with open(output, "w") as f:
        f.write(rendered_html)
    webbrowser.open(f"file://{output.resolve()}")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("path", type=Path, help="Path to the benchmark result directory")
    parser.add_argument("output", type=Path, help="Path to the benchmark result directory")
    args = parser.parse_args()
    generate_list(args.path, args.output)
